from typing import AsyncIterable

from aett.eventstore import IManagePersistenceAsync, COMMITS, Commit
from aett.storage.asynchronous.s3 import AsyncS3Config


class AsyncPersistenceManagement(IManagePersistenceAsync):
    def __init__(self, s3_config: AsyncS3Config, folder_name=COMMITS):
        self._folder_name = folder_name
        self._s3_bucket = s3_config.bucket
        self.__config = s3_config

    async def initialize(self) -> None:
        try:
            async with self.__config.to_client() as client:
                await client.create_bucket(Bucket=self._s3_bucket)
        except:
            pass

    async def drop(self) -> None:
        async with self.__config.to_client() as client:
            await client.delete_bucket(Bucket=self._s3_bucket)

    async def purge(self, tenant_id: str) -> None:
        async with self.__config.to_client() as client:
            response = await client.list_objects_v2(
                Bucket=self._s3_bucket, Prefix=f"{self._folder_name}/{tenant_id}/"
            )

            for o in response["Contents"]:
                await client.delete_object(Bucket=self._s3_bucket, Key=o["Key"])

    async def get_from(self, checkpoint: int) -> AsyncIterable[Commit]:
        raise NotImplementedError()



import datetime
from abc import ABC
from typing import Optional

from pydantic import BaseModel, Field


class BaseCommand(ABC, BaseModel):
    aggregate_id: str = Field(
        description="The id of the aggregate to apply the command to."
    )
    version: int = Field(description="The version of the aggregate to target")
    timestamp: datetime.datetime = Field(
        description="Gets the point in time at which the command was generated"
    )
    correlation_id: Optional[str] = Field(
        default=None, description="Gets the correlation id of the command."
    )



import boto3
import botocore.client
from botocore.client import BaseClient
import inspect


class S3Config:
    def __init__(
        self,
        bucket: str,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        profile_name: str | None = None,
        region: str = "us-east-1",
        endpoint_url: str | None = None,
        use_tls: bool = True,
    ):
        """
        Defines the configuration for the S3 client.
        If a profile name is provided, the access key id and secret access are disregarded and the profile credentials
        are used.

        :param bucket: The name of the bucket
        :param aws_access_key_id: The AWS access key id
        :param aws_secret_access_key: The AWS secret access key
        :param aws_session_token: The AWS session token
        :param region: The AWS region
        :param endpoint_url: The endpoint URL
        :param use_tls: Whether to use TLS
        :param profile_name: The profile name
        """
        self._aws_session_token = aws_session_token
        self._aws_secret_access_key = aws_secret_access_key
        self._aws_access_key_id = aws_access_key_id
        self._use_tls = use_tls
        self.bucket = bucket
        self._region = region
        self._endpoint_url = endpoint_url
        self._profile_name = profile_name

    def to_client(self) -> BaseClient:
        session = boto3.Session(
            profile_name=self._profile_name,
            aws_access_key_id=self._aws_access_key_id,
            aws_secret_access_key=self._aws_secret_access_key,
            aws_session_token=self._aws_session_token,
        )
        return session.client(
            service_name="s3",
            region_name=self._region,
            endpoint_url=self._endpoint_url,
            verify=self._use_tls,
        )



from abc import ABC, abstractmethod
from typing import Iterable

from aett.eventstore import Commit


class IManagePersistence(ABC):
    @abstractmethod
    def initialize(self) -> None:
        """
        Initializes the persistence mechanism.
        """
        pass

    @abstractmethod
    def drop(self) -> None:
        """
        Drops the persistence mechanism.
        """
        pass

    @abstractmethod
    def purge(self, tenant_id: str) -> None:
        """
        Purges the persistence mechanism.

        :param tenant_id: The value which uniquely identifies the tenant to be purged.
        """
        pass

    @abstractmethod
    def get_from(self, checkpoint: int) -> Iterable[Commit]:
        """
        Gets the commits from the checkpoint.
        :param checkpoint: The checkpoint to start from.
        :return: The commits from the checkpoint.
        """
        pass



from abc import ABC
from typing import List, Dict, Any

from aett.domain.constants import TCommand, UNDISPATCHEDMESSAGES
from aett.eventstore import EventMessage, BaseEvent


class Saga(ABC):
    """
    A saga is a long-running process that coordinates multiple services to achieve a goal.
    The saga base class requires implementors to provide a method to apply events during state transition.

    In addition to this, the aggregate base class provides a method to raise events, but the concrete application
    of the event relies on multiple dispatch to call the correct apply method in the subclass.
    """

    def __init__(self, saga_id: str, commit_sequence: int):
        """
        Initialize the saga

        :param saga_id: The id of the saga
        :param commit_sequence: The commit sequence number which the saga was built from
        """
        self._id = saga_id
        self._commit_sequence = commit_sequence
        self._version = 0
        self.uncommitted: List[EventMessage] = []
        self._headers: Dict[str, Any] = {}

    @property
    def id(self) -> str:
        """
        Gets the id of the saga
        """
        return self._id

    @property
    def version(self) -> int:
        """
        Gets the version of the saga
        """
        return self._version

    @property
    def commit_sequence(self) -> int:
        """
        Gets the commit sequence number of the saga
        """
        return self._commit_sequence

    @property
    def headers(self) -> Dict[str, Any]:
        """
        Gets the metadata headers of the saga
        """
        return self._headers

    def transition(self, event: BaseEvent) -> None:
        """
        Transitions the saga to the next state based on the event
        :param event: The trigger event
        :return: None
        """
        # Use multiple dispatch to call the correct apply method
        self._apply(event)
        self.uncommitted.append(EventMessage(body=event, headers=self._headers))
        self._version += 1

    def dispatch(self, command: TCommand) -> None:
        """
        Adds a command to the stream to be dispatched when the saga is committed
        :param command: The command to dispatch
        :return: None
        """
        from aett.eventstore import Topic

        topic_header = Topic.get(type(command))
        self._headers[f"{UNDISPATCHEDMESSAGES}.{len(self._headers)}"] = EventMessage(
            body=command, headers={"topic": topic_header}
        )



import boto3


def _get_resource(
    profile_name: str | None = None,
    aws_access_key_id: str | None = None,
    aws_secret_access_key: str | None = None,
    aws_session_token: str | None = None,
    region: str = "eu-central-1",
    port: int = 8000,
):
    session = boto3.Session(
        profile_name=profile_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
    )
    return session.resource(
        "dynamodb",
        region_name=region,
        endpoint_url=f"http://localhost:{port}" if region == "localhost" else None,
    )



from aett.domain.aggregate import Aggregate
from aett.domain.aggregate_repository import AggregateRepository
from aett.domain.default_aggregate_repository import DefaultAggregateRepository
from aett.domain.async_default_aggregate_repository import (
    AsyncDefaultAggregateRepository,
)
from aett.domain.conflict_delegate import ConflictDelegate
from aett.domain.conflict_detector import ConflictDetector
from aett.domain.saga import Saga
from aett.domain.saga_repository import SagaRepository
from aett.domain.async_saga_repository import AsyncSagaRepository
from aett.domain.constants import TMemento, TCommitted, TUncommitted
from aett.domain.conflicting_commit_exception import ConflictingCommitException
from aett.domain.duplicate_commit_exception import DuplicateCommitException
from aett.domain.non_conflicting_commit_exception import NonConflictingCommitException



import typing

import asyncpg
from pydantic_core import from_json, to_json

from aett.eventstore import IAccessSnapshotsAsync, SNAPSHOTS, MAX_INT, Snapshot


class AsyncSnapshotStore(IAccessSnapshotsAsync):
    def __init__(self, connection_string: str, table_name: str = SNAPSHOTS):
        self._connection_string: str = connection_string
        self._table_name = table_name

    async def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        try:
            connection = await asyncpg.connect(self._connection_string)
            results = await connection.fetch(
                f"""SELECT *
          FROM {self._table_name}
         WHERE TenantId = $1
           AND StreamId = $2
           AND StreamRevision <= $3
         ORDER BY StreamRevision DESC
         LIMIT 1;""",
                tenant_id,
                stream_id,
                max_revision,
            )
            if not results:
                return None
            item = results[0]

            return Snapshot(
                tenant_id=item[0],
                stream_id=item[1],
                stream_revision=int(item[2]),
                commit_sequence=int(item[3]),
                payload=from_json(item[4]),
                headers=dict(from_json(item[5])),
            )
        except Exception as e:
            raise Exception(
                f"Failed to get snapshot for stream {stream_id} with error {e}"
            )

    async def add(
        self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None
    ):
        if headers is None:
            headers = {}
        try:
            connection = await asyncpg.connect(self._connection_string)
            await connection.execute(
                f"""INSERT INTO {self._table_name} ( TenantId, StreamId, StreamRevision, CommitSequence, Payload, Headers) VALUES ($1, $2, $3, $4, $5, $6);""",
                snapshot.tenant_id,
                snapshot.stream_id,
                snapshot.stream_revision,
                snapshot.commit_sequence,
                to_json(snapshot.payload),
                to_json(headers),
            )
            await connection.close()
        except Exception as e:
            raise Exception(
                f"Failed to add snapshot for stream {snapshot.stream_id} with error {e}"
            )



import datetime
from abc import ABC, abstractmethod
from typing import TypeVar, Type, Dict

from aett.domain import Aggregate


class AsyncAggregateRepository(ABC):
    """
    Defines the abstract interface for an aggregate repository.
    The repository is responsible for loading and saving aggregates to the event store,
    typically using the ICommitEvents interface.
    """

    TAggregate = TypeVar("TAggregate", bound=Aggregate)

    @abstractmethod
    async def get(
        self, cls: Type[TAggregate], stream_id: str, max_version: int = 2**32
    ) -> TAggregate:
        """
        Gets the aggregate with the specified stream id and type

        :param cls: The type of the aggregate
        :param stream_id: The id of the stream to load
        :param max_version: The max aggregate version to load.
        """
        pass

    @abstractmethod
    async def get_to(
        self,
        cls: Type[TAggregate],
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> TAggregate:
        """
        Gets the aggregate with the specified stream id and type

        :param cls: The type of the aggregate
        :param stream_id: The id of the stream to load
        :param max_time: The max aggregate timestamp to load.
        """
        pass

    @abstractmethod
    async def save(
        self, aggregate: TAggregate, headers: Dict[str, str] | None = None
    ) -> None:
        """
        Save the aggregate to the repository.

        The call to save should be wrapped in a try-except block as concurrent modifications can cause a conflict with
        events committed from a different source. A ConflictingCommitException will be thrown by the storage layer if
        an attempt is made to save an aggregate with a version that is lower than the current version in the store and
        the uncommitted events conflict with the committed events.A NonConflictingCommitException will be thrown if the
        uncommitted events do not conflict with the committed events. In this case it should be safe to retry the
        operation.

        :param aggregate: The aggregate to save.
        :param headers: The headers to assign to the commit.
        """
        pass

    @abstractmethod
    async def snapshot(
        self,
        cls: Type[TAggregate],
        stream_id: str,
        version: int,
        headers: Dict[str, str],
    ) -> None:
        """
        Generates a snapshot of the aggregate at the specified version.

        :param cls: The type of the aggregate
        :param stream_id: The id of the aggregate to snapshot
        :param version: The version of the aggregate to snapshot
        :param headers: The headers to assign to the snapshot
        """
        pass

    @abstractmethod
    async def snapshot_at(
        self,
        cls: Type[TAggregate],
        stream_id: str,
        cut_off: datetime.datetime,
        headers: Dict[str, str],
    ) -> None:
        """
        Generates a snapshot of the aggregate at the specified time point.

        :param cls: The type of the aggregate
        :param stream_id: The id of the aggregate to snapshot
        :param cut_off: The time point of the aggregate to snapshot
        :param headers: The headers to assign to the snapshot
        """
        pass



import typing
from abc import ABC

from pydantic import BaseModel

from aett.eventstore.constants import T


class Memento(ABC, BaseModel, typing.Generic[T]):
    id: str
    """
    Gets the id of the aggregate which generated the memento.
    """

    version: int
    """
    Gets the version of the aggregate which generated the memento.
    """

    payload: T
    """
    Gets the state of the aggregate at the time the memento was taken.
    """



class Topic(object):
    """
    Represents the topic of an event message.
    Should be used as a decorator on a class to indicate the topic of the event which will help with type deserialization.
    """

    def __init__(self, topic: str):
        self.topic = topic

    def __call__(self, cls):
        cls.__topic__ = self.topic
        return cls

    @staticmethod
    def get(cls: type) -> str:
        return cls.__topic__ if hasattr(cls, "__topic__") else cls.__name__



from abc import ABC, abstractmethod
from typing import TypeVar, Type, Dict

from aett.domain.saga import Saga


class SagaRepository(ABC):
    """
    Defines the abstract interface for a saga repository.
    The repository is responsible for loading and saving sagas to the event store,
    typically using the ICommitEvents interface.
    """

    TSaga = TypeVar("TSaga", bound=Saga)

    @abstractmethod
    def get(self, cls: Type[TSaga], stream_id: str) -> TSaga:
        """
        Gets the saga with the specified stream id and type at the latest version.
        """
        pass

    @abstractmethod
    def save(self, saga: Saga) -> None:
        """
        Save the saga to the repository.
        """
        pass



class NonConflictingCommitException(Exception):
    """
    Exception raised when a non-conflicting commit is detected.
    """

    def __init__(self, message: str):
        super().__init__(message)



from abc import ABC, abstractmethod
from typing import Generic

from aett.domain.constants import TUncommitted, TCommitted


class ConflictDelegate(ABC, Generic[TUncommitted, TCommitted]):
    """
    A conflict delegate is a class that can detect conflicts between two events.
    """

    @abstractmethod
    def detect(self, uncommitted: TUncommitted, committed: TCommitted) -> bool:
        """
        Detects if the uncommitted event conflicts with the committed event. The delegate should return True if an event
        is incompatible with a previously persisted event.

        If the delegate returns False then it is assumed that the later event is compatible with the previously
        persisted.
        """
        pass



import datetime
import typing
from typing import Iterable
from uuid import UUID

from boto3.dynamodb.conditions import Key, Attr
from pydantic_core import from_json, to_json

from aett.domain import (
    ConflictDetector,
    DuplicateCommitException,
    ConflictingCommitException,
    NonConflictingCommitException,
)
from aett.storage.synchronous.dynamodb import _get_resource
from aett.eventstore import TopicMap, COMMITS, MAX_INT, Commit, EventMessage, BaseEvent
from aett.eventstore.i_commit_events import ICommitEvents


class CommitStore(ICommitEvents):
    def __init__(
        self,
        topic_map: TopicMap,
        conflict_detector: ConflictDetector | None = None,
        table_name: str = COMMITS,
        region: str = "eu-central-1",
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        port: int = 8000,
    ):
        self._topic_map = topic_map
        self._table_name = table_name
        self._region = region
        self._dynamodb = _get_resource(
            profile_name=profile_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region=region,
            port=port,
        )
        self._table = self._dynamodb.Table(table_name)
        self._conflict_detector: ConflictDetector = (
            conflict_detector if conflict_detector is not None else ConflictDetector()
        )

    def get(
        self,
        tenant_id: str,
        stream_id: str,
        min_revision: int = 0,
        max_revision: int = MAX_INT,
    ) -> typing.Iterable[Commit]:
        max_revision = MAX_INT if max_revision >= MAX_INT else max_revision + 1
        min_revision = 0 if min_revision < 0 else min_revision
        query_response = self._table.query(
            TableName=self._table_name,
            IndexName="RevisionIndex",
            ConsistentRead=True,
            ProjectionExpression="TenantId,StreamId,StreamRevision,CommitId,CommitSequence,CommitStamp,Headers,Events",
            KeyConditionExpression=(
                Key("TenantAndStream").eq(f"{tenant_id}{stream_id}")
                & Key("StreamRevision").between(min_revision, max_revision)
            ),
            ScanIndexForward=True,
        )
        items = query_response["Items"]
        for item in items:
            yield self._item_to_commit(item)

    def get_to(
        self,
        tenant_id: str,
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> Iterable[Commit]:
        query_response = self._table.scan(
            IndexName="CommitStampIndex",
            ConsistentRead=True,
            Select="ALL_ATTRIBUTES",
            FilterExpression=(
                Key("TenantAndStream").eq(f"{tenant_id}{stream_id}")
                & Attr("CommitStamp").lte(int(max_time.timestamp()))
            ),
        )
        items = query_response["Items"]
        for item in items:
            if item["CommitStamp"] > max_time.timestamp():
                break
            yield self._item_to_commit(item)

    def get_all_to(
        self, tenant_id: str, max_time: datetime.datetime = datetime.datetime.max
    ) -> Iterable[Commit]:
        query_response = self._table.scan(
            IndexName="CommitStampIndex",
            ConsistentRead=True,
            Select="ALL_ATTRIBUTES",
            ProjectionExpression="CommitStamp",
            FilterExpression=(
                Key("TenantAndStream").begins_with(f"{tenant_id}")
                & Attr("CommitStamp").lte(int(max_time.timestamp()))
            ),
        )
        items = query_response["Items"]
        for item in items:
            if item["CommitStamp"] > max_time.timestamp():
                break
            yield self._item_to_commit(item)

    def _item_to_commit(self, item: dict) -> Commit:
        return Commit(
            tenant_id=item["TenantId"],
            stream_id=item["StreamId"],
            stream_revision=int(item["StreamRevision"]),
            commit_id=UUID(item["CommitId"]),
            commit_sequence=int(item["CommitSequence"]),
            commit_stamp=datetime.datetime.fromtimestamp(
                int(item["CommitStamp"]), datetime.timezone.utc
            ),
            headers=from_json(bytes(item["Headers"])),
            events=[
                EventMessage.from_dict(e, self._topic_map)
                for e in from_json(bytes(item["Events"]))
            ],
            checkpoint_token=0,
        )

    def commit(self, commit: Commit):
        try:
            item = {
                "TenantAndStream": f"{commit.tenant_id}{commit.stream_id}",
                "TenantId": commit.tenant_id,
                "StreamId": commit.stream_id,
                "StreamRevision": commit.stream_revision,
                "CommitId": str(commit.commit_id),
                "CommitSequence": commit.commit_sequence,
                "CommitStamp": int(commit.commit_stamp.timestamp()),
                "Headers": to_json(commit.headers),
                "Events": to_json([e.to_json() for e in commit.events]),
            }
            _ = self._table.put_item(
                TableName=self._table_name,
                Item=item,
                ReturnValues="NONE",
                ReturnValuesOnConditionCheckFailure="NONE",
                ConditionExpression="attribute_not_exists(TenantAndStream) AND attribute_not_exists(CommitSequence)",
            )
        except Exception as e:
            if e.__class__.__name__ == "ConditionalCheckFailedException":
                if self._detect_duplicate(
                    commit.commit_id,
                    commit.tenant_id,
                    commit.stream_id,
                    commit.commit_sequence,
                ):
                    raise DuplicateCommitException("Duplicate commit detected")
                else:
                    self._raise_conflict(commit)
            else:
                raise e

    def _raise_conflict(self, commit: Commit):
        if self._detect_conflicts(commit=commit):
            raise ConflictingCommitException(
                f"Conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
            )
        else:
            raise NonConflictingCommitException(
                f"Non-conflicting version conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
            )

    def _detect_duplicate(
        self, commit_id: UUID, tenant_id: str, stream_id: str, commit_sequence: int
    ) -> bool:
        duplicate_check = self._table.query(
            TableName=self._table_name,
            ConsistentRead=True,
            ScanIndexForward=False,
            Limit=1,
            Select="SPECIFIC_ATTRIBUTES",
            ProjectionExpression="CommitId",
            KeyConditionExpression=(
                Key("TenantAndStream").eq(f"{tenant_id}{stream_id}")
                & Key("CommitSequence").eq(commit_sequence)
            ),
        )
        items = duplicate_check["Items"]
        return items[0]["CommitId"] == str(commit_id)

    def _detect_conflicts(self, commit: Commit) -> bool:
        if commit.commit_sequence == 0:
            return False
        previous_commits = self.get(
            commit.tenant_id,
            commit.stream_id,
            commit.commit_sequence - 1,
            commit.commit_sequence,
        )
        for previous_commit in previous_commits:
            if self._conflict_detector.conflicts_with(
                list(map(self._get_body, commit.events)),
                list(map(self._get_body, previous_commit.events)),
            ):
                return True
        return False

    @staticmethod
    def _get_body(em: EventMessage) -> BaseEvent:
        body: BaseEvent = em.body
        return body



import datetime
import logging
import uuid
import typing

from aett.domain.saga_repository import SagaRepository, Saga
from aett.eventstore import ICommitEvents, Commit


class DefaultSagaRepository(SagaRepository):
    def __init__(
        self, tenant_id: str, store: ICommitEvents, logger: logging.Logger | None = None
    ):
        """
        Initialize the default saga repository.

        :param tenant_id: The tenant id of the repository instance
        :param store: The event store to use
        """
        self._tenant_id = tenant_id
        self._store = store
        self._logger = (
            logger
            if logger is not None
            else logging.getLogger(DefaultSagaRepository.__name__)
        )

    def get(
        self, cls: typing.Type[SagaRepository.TSaga], stream_id: str
    ) -> SagaRepository.TSaga:
        self._logger.debug(f"Getting saga {cls.__name__} with id {stream_id}")
        ait = self._store.get(self._tenant_id, stream_id)
        commits = [x for x in ait]
        commit_sequence = commits[-1].commit_sequence if len(commits) > 0 else 0
        saga: Saga = cls(stream_id, commit_sequence)
        for commit in commits:
            for event in commit.events:
                saga.transition(event.body)
        saga.uncommitted.clear()
        saga.headers.clear()
        return saga

    def save(self, saga: Saga) -> None:
        self._logger.debug(f"Saving saga {saga.id} at version {saga.version}")
        commit = Commit(
            tenant_id=self._tenant_id,
            stream_id=saga.id,
            stream_revision=saga.version,
            commit_id=uuid.uuid4(),
            commit_sequence=saga.commit_sequence + 1,
            commit_stamp=datetime.datetime.now(datetime.timezone.utc),
            headers=dict(saga.headers),
            events=list(saga.uncommitted),
            checkpoint_token=0,
        )
        self._store.commit(commit=commit)
        self._logger.debug(f"Saved saga {saga.id}")
        saga.uncommitted.clear()
        saga.headers.clear()



import datetime
from abc import ABC
from typing import Optional

from pydantic import BaseModel, Field


class BaseEvent(ABC, BaseModel):
    """
    Represents a single event which has occurred outside the application domain.
    """

    source: str = Field(
        description="Gets the value which uniquely identifies the source of the event."
    )

    timestamp: datetime.datetime = Field(
        description="Gets the point in time at which the event was generated."
    )

    correlation_id: Optional[str] = Field(
        default=None, description="Gets the correlation id of the event."
    )



import logging
from typing import AsyncIterable

import asyncpg
from asyncpg import Connection

from aett.eventstore import (
    IManagePersistenceAsync,
    TopicMap,
    COMMITS,
    SNAPSHOTS,
    Commit,
)
from aett.storage.asynchronous.postgresql import _item_to_commit


class AsyncPersistenceManagement(IManagePersistenceAsync):
    def __init__(
        self,
        connection_string: str,
        topic_map: TopicMap,
        commits_table_name: str = COMMITS,
        snapshots_table_name: str = SNAPSHOTS,
    ):
        self._connection_string: str = connection_string
        self._topic_map = topic_map
        self._commits_table_name = commits_table_name
        self._snapshots_table_name = snapshots_table_name

    async def initialize(self):
        try:
            connection: Connection = await asyncpg.connect(self._connection_string)
            await connection.execute(f"""CREATE TABLE {self._commits_table_name}
        (
            TenantId varchar(64) NOT NULL,
            StreamId char(64) NOT NULL,
            StreamIdOriginal varchar(1000) NOT NULL,
            StreamRevision int NOT NULL CHECK (StreamRevision > 0),
            Items smallint NOT NULL CHECK (Items > 0),
            CommitId uuid NOT NULL,
            CommitSequence int NOT NULL CHECK (CommitSequence > 0),
            CommitStamp timestamp with time zone NOT NULL,
            CheckpointNumber BIGSERIAL NOT NULL,
            Headers bytea NULL,
            Payload bytea NOT NULL,
            CONSTRAINT PK_Commits PRIMARY KEY (CheckpointNumber)
        );
        CREATE UNIQUE INDEX IX_Commits_CommitSequence ON {self._commits_table_name} (TenantId, StreamId, CommitSequence);
        CREATE UNIQUE INDEX IX_Commits_CommitId ON {self._commits_table_name} (TenantId, StreamId, CommitId);
        CREATE UNIQUE INDEX IX_Commits_Revisions ON {self._commits_table_name} (TenantId, StreamId, StreamRevision, Items);
        CREATE INDEX IX_Commits_Stamp ON {self._commits_table_name} (CommitStamp);
        
        CREATE TABLE {self._snapshots_table_name}
        (
            TenantId varchar(40) NOT NULL,
            StreamId char(40) NOT NULL,
            StreamRevision int NOT NULL CHECK (StreamRevision > 0),
            CommitSequence int NOT NULL CHECK (CommitSequence > 0),
            Payload bytea NOT NULL,
            Headers bytea NOT NULL,
            CONSTRAINT PK_Snapshots PRIMARY KEY (TenantId, StreamId, StreamRevision)
        );""")
        except Exception as e:
            logging.error(
                f"Failed to initialize persistence with error {e}", exc_info=True
            )

    async def drop(self):
        connection = await asyncpg.connect(self._connection_string)
        await connection.execute(
            f"""DROP TABLE {self._snapshots_table_name};DROP TABLE {self._commits_table_name};"""
        )

    async def purge(self, tenant_id: str):
        connection: Connection = await asyncpg.connect(self._connection_string)
        await connection.execute(
            f"""DELETE FROM {self._commits_table_name} WHERE TenantId = %s;""",
            tenant_id,
        )
        await connection.execute(
            f"""DELETE FROM {self._snapshots_table_name} WHERE TenantId = %s;""",
            tenant_id,
        )
        await connection.close()

    async def get_from(self, checkpoint: int) -> AsyncIterable[Commit]:
        connection: Connection = await asyncpg.connect(self._connection_string)
        fetchall = await connection.execute(
            f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                                      FROM {self._commits_table_name}
                                     WHERE CommitStamp >= %s
                                     ORDER BY CheckpointNumber;""",
            (checkpoint,),
        )
        for doc in fetchall:
            yield _item_to_commit(doc, self._topic_map)



from aett.eventstore.base_command import BaseCommand
from aett.eventstore.base_event import BaseEvent
from aett.eventstore.commit import Commit
from aett.eventstore.domain_event import DomainEvent
from aett.eventstore.event_message import EventMessage
from aett.eventstore.i_access_snapshots import IAccessSnapshots
from aett.eventstore.i_access_snapshots_async import IAccessSnapshotsAsync
from aett.eventstore.i_commit_events import ICommitEvents
from aett.eventstore.i_commit_events_async import ICommitEventsAsync
from aett.eventstore.i_manage_persistence import IManagePersistence
from aett.eventstore.i_manage_persistence_async import IManagePersistenceAsync
from aett.eventstore.memento import Memento
from aett.eventstore.snapshot import Snapshot
from aett.eventstore.stream_head import StreamHead
from aett.eventstore.topic import Topic
from aett.eventstore.topic_map import TopicMap
from aett.eventstore.constants import MAX_INT, COMMITS, SNAPSHOTS



import typing
import datetime
from uuid import UUID

from aett.domain import (
    ConflictDetector,
    ConflictingCommitException,
    NonConflictingCommitException,
)
from aett.eventstore import (
    Commit,
    TopicMap,
    ICommitEventsAsync,
    COMMITS,
    MAX_INT,
    EventMessage,
    BaseEvent,
)
from pymongo import ASCENDING
from pymongo.errors import DuplicateKeyError
from pydantic_core import to_json
from pymongo.asynchronous import database, collection, cursor
from pymongo.asynchronous.cursor import AsyncCursor

from aett.storage.asynchronous.mongodb.mapping import _doc_to_commit


class AsyncCommitStore(ICommitEventsAsync):
    def __init__(
        self,
        db: database.AsyncDatabase,
        topic_map: TopicMap,
        conflict_detector: ConflictDetector | None = None,
        table_name=COMMITS,
    ):
        self._topic_map = topic_map
        self._collection: collection.AsyncCollection = db.get_collection(table_name)
        self._counters_collection: collection.AsyncCollection = db.get_collection(
            "counters"
        )
        self._conflict_detector = (
            conflict_detector if conflict_detector is not None else ConflictDetector()
        )

    async def get(
        self,
        tenant_id: str,
        stream_id: str,
        min_revision: int = 0,
        max_revision: int = MAX_INT,
    ) -> typing.AsyncIterable[Commit]:
        max_revision = MAX_INT if max_revision >= MAX_INT else max_revision + 1
        min_revision = 0 if min_revision < 0 else min_revision
        filters: typing.Dict[str, typing.Any] = {
            "TenantId": tenant_id,
            "StreamId": stream_id,
        }
        if min_revision > 0:
            filters["StreamRevision"] = {"$gte": min_revision}
        if max_revision < MAX_INT:
            if "StreamRevision" in filters:
                filters["StreamRevision"]["$lte"] = max_revision
            else:
                filters["StreamRevision"] = {"$lte": max_revision}

        query_response: AsyncCursor = self._collection.find({"$and": [filters]})
        async for doc in query_response.sort("CheckpointToken", direction=ASCENDING):
            yield _doc_to_commit(doc, self._topic_map)

    async def get_to(
        self,
        tenant_id: str,
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> typing.AsyncIterable[Commit]:
        filters = {
            "TenantId": tenant_id,
            "StreamId": stream_id,
            "CommitStamp": {"$lte": int(max_time.timestamp())},
        }

        query_response: cursor.AsyncCursor = self._collection.find({"$and": [filters]})
        async for doc in query_response.sort("CheckpointToken", direction=ASCENDING):
            yield _doc_to_commit(doc, self._topic_map)

    async def get_all_to(
        self, tenant_id: str, max_time: datetime.datetime = datetime.datetime.max
    ) -> typing.AsyncIterable[Commit]:
        filters = {
            "TenantId": tenant_id,
            "CommitStamp": {"$lte": int(max_time.timestamp())},
        }

        query_response: cursor.AsyncCursor = self._collection.find({"$and": [filters]})
        async for doc in query_response.sort("CheckpointToken", direction=ASCENDING):
            yield _doc_to_commit(doc, self._topic_map)

    async def commit(self, commit: Commit) -> Commit:
        try:
            seq_ = await self._counters_collection.find_one_and_update(
                filter={"_id": "CheckpointToken"}, update={"$inc": {"seq": 1}}
            )
            ret = seq_.get("seq")
            doc = {
                "TenantId": commit.tenant_id,
                "StreamId": commit.stream_id,
                "StreamRevision": commit.stream_revision,
                "CommitId": str(commit.commit_id),
                "CommitSequence": commit.commit_sequence,
                "CommitStamp": int(
                    datetime.datetime.now(datetime.timezone.utc).timestamp()
                ),
                "Headers": to_json(commit.headers),
                "Events": to_json([e.to_json() for e in commit.events]),
                "CheckpointToken": int(ret),
            }
            _ = await self._collection.insert_one(doc)
            return Commit(
                tenant_id=commit.tenant_id,
                stream_id=commit.stream_id,
                stream_revision=commit.stream_revision,
                commit_id=commit.commit_id,
                commit_sequence=commit.commit_sequence,
                commit_stamp=commit.commit_stamp,
                headers=commit.headers,
                events=commit.events,
                checkpoint_token=ret,
            )
        except Exception as e:
            if isinstance(e, DuplicateKeyError):
                if await self._detect_duplicate(
                    commit.commit_id,
                    commit.tenant_id,
                    commit.stream_id,
                    commit.commit_sequence,
                ):
                    raise Exception(
                        f"Commit {commit.commit_id} already exists in stream {commit.stream_id}"
                    )
                else:
                    conflicts, revision = await self._detect_conflicts(commit=commit)
                    if conflicts:
                        raise ConflictingCommitException(
                            f"Conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                        )
                    else:
                        raise NonConflictingCommitException(
                            f"Non-conflicting version conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                        )
            else:
                raise Exception(
                    f"Failed to commit event to stream {commit.stream_id} with status code {e}"
                )

    async def _detect_duplicate(
        self, commit_id: UUID, tenant_id: str, stream_id: str, commit_sequence: int
    ) -> bool:
        duplicate_check = await self._collection.find_one(
            {
                "TenantId": tenant_id,
                "StreamId": stream_id,
                "CommitSequence": commit_sequence,
            }
        )
        if not duplicate_check:
            return False
        s = str(duplicate_check.get("CommitId"))
        return s == str(commit_id)

    async def _detect_conflicts(self, commit: Commit) -> typing.Tuple[bool, int]:
        filters = {
            "TenantId": commit.tenant_id,
            "StreamId": commit.stream_id,
            "CommitSequence": {"$lte": commit.commit_sequence},
        }
        query_response: cursor.AsyncCursor = self._collection.find(
            {"$and": [filters]}
        ).sort("CheckpointToken", direction=ASCENDING)

        latest_revision = 0
        async for doc in query_response:
            c = _doc_to_commit(doc, self._topic_map)
            if self._conflict_detector.conflicts_with(
                list(map(self._get_body, commit.events)),
                list(map(self._get_body, c.events)),
            ):
                return True, -1
            i = int(doc["StreamRevision"])
            if i > latest_revision:
                latest_revision = i
        return False, latest_revision

    @staticmethod
    def _get_body(e: EventMessage) -> BaseEvent:
        return e.body



import inspect
from typing import Any, Dict, List, Self

from aett.eventstore.topic import Topic


class TopicMap:
    """
    Represents a map of topics to event classes.
    """

    def __init__(self):
        self.__topics: Dict[str, type] = {}

    def add(self, topic: str, cls: type) -> Self:
        """
        Adds the topic and class to the map.
        :param topic: The topic of the event.
        :param cls: The class of the event.
        """
        self.__topics[topic] = cls
        return self

    def register(self, instance: Any) -> Self:
        t = instance if isinstance(instance, type) else type(instance)
        topic = Topic.get(t)
        if topic not in self.__topics:
            self.add(topic, t)

        return self

    def register_module(self, module: object) -> Self:
        """
        Registers all the classes in the module.
        """
        for _, o in inspect.getmembers(module, inspect.isclass):
            if inspect.isclass(o):
                self.register(o)
            if inspect.ismodule(o):
                self.register_module(o)
        return self

    def get(self, topic: str) -> type | None:
        """
        Gets the class of the event given the topic.
        :param topic: The topic of the event.
        :return: The class of the event.
        """
        return self.__topics.get(topic, None)

    def get_all(self) -> List[str]:
        """
        Gets all the topics and their corresponding classes in the map.
        :return: A dictionary of all the topics and their classes.
        """
        return list(self.__topics.keys())

    def get_all_types(self) -> List[type]:
        """
        Gets all the types in the map.
        :return: A list of all the types in the map.
        """
        return list(self.__topics.values())



import datetime
from typing import Dict, List, Annotated
from uuid import UUID

from pydantic import BaseModel, Field, PlainSerializer, BeforeValidator

from aett.eventstore.event_message import EventMessage


def _convert_event_list(event_list: List[EventMessage]) -> List[dict]:
    return [x.to_json() for x in event_list]


class Commit(BaseModel):
    """
    Represents a series of events which have been fully committed as a single unit
    and which apply to the stream indicated.
    """

    tenant_id: str
    """
    Gets or sets the value which identifies tenant to which the stream and the commit belongs.
    """

    stream_id: str
    """
    Gets the value which uniquely identifies the stream to which the commit belongs.
    """

    stream_revision: int
    """
    Gets the value which indicates the revision of the most recent event in the stream to which this commit applies.
    """

    commit_id: UUID
    """
    Gets the value which uniquely identifies the commit within the stream.
    """

    commit_sequence: int
    """
    Gets the value which indicates the sequence (or position) in the stream to which this commit applies.
    """

    commit_stamp: datetime.datetime
    """
    Gets the point in time at which the commit was persisted.
    """

    headers: Dict[str, object]
    """
    Gets the metadata which provides additional, unstructured information about this commit.
    """

    events: Annotated[
        List[EventMessage],
        PlainSerializer(
            func=_convert_event_list, return_type=List[dict], when_used="json"
        ),
    ] = Field(default_factory=list)
    """
    Gets the collection of event messages to be committed as a single unit.
    """

    checkpoint_token: int
    """
    The checkpoint that represents the storage level order.
    """



from aett.eventstore import BaseEvent


class DomainEvent(BaseEvent):
    """
    Represents a single event which has occurred within the domain.
    """

    version: int
    """
    Gets the version of the aggregate which generated the event.
    """



from abc import ABC, abstractmethod
from typing import Optional, Dict

from aett.eventstore.snapshot import Snapshot


class IAccessSnapshotsAsync(ABC):
    """
    Indicates the ability to get and add snapshots.
    """

    @abstractmethod
    async def get(
        self, tenant_id: str, stream_id: str, max_revision: int
    ) -> Optional[Snapshot]:
        """
        Gets the snapshot at the revision indicated or the most recent snapshot below that revision.

        :param tenant_id: The value which uniquely identifies the bucket to which the stream and the snapshot belong.
        :param stream_id: The stream for which the snapshot should be returned.
        :param max_revision: The maximum revision possible for the desired snapshot.
        :return: If found, returns the snapshot for the stream indicated; otherwise null.
        :raises StorageException:
        :raises StorageUnavailableException:
        """
        pass

    @abstractmethod
    async def add(
        self, snapshot: Snapshot, headers: Dict[str, str] | None = None
    ) -> None:
        """
        Adds the snapshot provided to the stream indicated. Using a snapshotId of Guid.Empty will always persist the
        snapshot.

        :param snapshot: The snapshot to save.
        :param headers: The metadata to assign to the snapshot.
        :raises StorageException:
        :raises StorageUnavailableException:
        """
        pass



from pydantic_core import from_json

from aett.eventstore import TopicMap, Commit, EventMessage


def _item_to_commit(item, topic_map: TopicMap):
    return Commit(
        tenant_id=item[0],
        stream_id=item[1],
        stream_revision=item[3],
        commit_id=item[4],
        commit_sequence=item[5],
        commit_stamp=item[6],
        headers=from_json(item[8]),
        events=[EventMessage.from_dict(e, topic_map) for e in from_json(item[9])],
        checkpoint_token=item[7],
    )



import datetime
from typing import AsyncIterable, Tuple
from uuid import UUID

import aiosqlite
from pydantic_core import to_json, from_json

from aett.domain import (
    ConflictDetector,
    DuplicateCommitException,
    ConflictingCommitException,
    NonConflictingCommitException,
)
from aett.eventstore import TopicMap, COMMITS, MAX_INT, Commit, EventMessage, BaseEvent
from aett.eventstore.i_commit_events_async import ICommitEventsAsync
from aett.storage.asynchronous.sqlite.functions import _item_to_commit


class AsyncCommitStore(ICommitEventsAsync):
    def __init__(
        self,
        connection_string: str,
        topic_map: TopicMap,
        conflict_detector: ConflictDetector | None = None,
        table_name=COMMITS,
    ):
        self._topic_map = topic_map
        self._connection_string = connection_string
        self._conflict_detector = (
            conflict_detector if conflict_detector is not None else ConflictDetector()
        )
        self._table_name = table_name

    async def get(
        self,
        tenant_id: str,
        stream_id: str,
        min_revision: int = 0,
        max_revision: int = MAX_INT,
    ) -> AsyncIterable[Commit]:
        max_revision = MAX_INT if max_revision >= MAX_INT else max_revision + 1
        min_revision = 0 if min_revision < 0 else min_revision
        async with aiosqlite.connect(self._connection_string) as connection:
            cur: aiosqlite.Cursor = await connection.cursor()
            await cur.execute(
                f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
          FROM {self._table_name}
         WHERE TenantId = ?
           AND StreamId = ?
           AND StreamRevision >= ?
           AND (StreamRevision - Items) < ?
           AND CommitSequence > ?
         ORDER BY CommitSequence;""",
                (tenant_id, stream_id, min_revision, max_revision, 0),
            )
            fetchall = await cur.fetchall()
            for doc in fetchall:
                yield _item_to_commit(doc, self._topic_map)

    async def get_to(
        self,
        tenant_id: str,
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> AsyncIterable[Commit]:
        async with aiosqlite.connect(self._connection_string) as connection:
            cur = await connection.cursor()
            await cur.execute(
                f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                      FROM {self._table_name}
                     WHERE TenantId = ?
                       AND StreamId = ?
                       AND CommitStamp <= ?
                     ORDER BY CommitSequence;""",
                (tenant_id, stream_id, max_time),
            )
            fetchall = await cur.fetchall()
            for doc in fetchall:
                yield _item_to_commit(doc, self._topic_map)

    async def get_all_to(
        self, tenant_id: str, max_time: datetime.datetime = datetime.datetime.max
    ) -> AsyncIterable[Commit]:
        async with aiosqlite.connect(self._connection_string) as connection:
            async with connection.cursor() as cur:
                await cur.execute(
                    f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                              FROM {self._table_name}
                             WHERE TenantId = ?
                              AND CommitStamp <= ?
                             ORDER BY CheckpointNumber;""",
                    (tenant_id, max_time),
                )
                fetchall = await cur.fetchall()
                for doc in fetchall:
                    yield _item_to_commit(doc, self._topic_map)

    async def commit(self, commit: Commit) -> Commit:
        try:
            async with aiosqlite.connect(self._connection_string) as connection:
                cur: aiosqlite.Cursor = await connection.cursor()
                json = to_json([e.to_json() for e in commit.events])
                await cur.execute(
                    f"""INSERT
              INTO {self._table_name}
                 ( TenantId, StreamId, StreamIdOriginal, CommitId, CommitSequence, StreamRevision, Items, CommitStamp, Headers, Payload )
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
            RETURNING CheckpointNumber;""",
                    (
                        commit.tenant_id,
                        commit.stream_id,
                        commit.stream_id,
                        commit.commit_id.bytes_le,
                        commit.commit_sequence,
                        commit.stream_revision,
                        len(commit.events),
                        commit.commit_stamp,
                        to_json(commit.headers),
                        json,
                    ),
                )
                checkpoint_number = cur.lastrowid
                await cur.close()
                await connection.commit()
                return Commit(
                    tenant_id=commit.tenant_id,
                    stream_id=commit.stream_id,
                    stream_revision=commit.stream_revision,
                    commit_id=commit.commit_id,
                    commit_sequence=commit.commit_sequence,
                    commit_stamp=commit.commit_stamp,
                    headers=commit.headers,
                    events=commit.events,
                    checkpoint_token=checkpoint_number,
                )

        except aiosqlite.IntegrityError:
            if await self._detect_duplicate(
                commit.commit_id, commit.tenant_id, commit.stream_id
            ):
                raise DuplicateCommitException(
                    f"Commit {commit.commit_id} already exists in stream {commit.stream_id}"
                )
            else:
                conflicts, revision = await self._detect_conflicts(commit=commit)
                if conflicts:
                    raise ConflictingCommitException(
                        f"Conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                    )
                else:
                    raise NonConflictingCommitException(
                        f"Non-conflicting version conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                    )

        except Exception as e:
            raise Exception(f"Failed to commit {commit.commit_id} with error {e}")

    async def _detect_duplicate(
        self, commit_id: UUID, tenant_id: str, stream_id: str
    ) -> bool:
        try:
            async with aiosqlite.connect(self._connection_string) as connection:
                cur: aiosqlite.Cursor = await connection.cursor()
                await cur.execute(
                    f"""SELECT COUNT(*)
                  FROM {self._table_name}
                 WHERE TenantId = ?
                   AND StreamId = ?
                   AND CommitId = ?;""",
                    (tenant_id, stream_id, str(commit_id)),
                )
                result = await cur.fetchone()
                await cur.close()
                await connection.commit()
                return result[0] > 0
        except Exception as e:
            raise Exception(
                f"Failed to detect duplicate commit {commit_id} with error {e}"
            )

    async def _detect_conflicts(self, commit: Commit) -> Tuple[bool, int]:
        async with aiosqlite.connect(self._connection_string) as connection:
            cur: aiosqlite.Cursor = await connection.cursor()
            await cur.execute(
                f"""SELECT StreamRevision, Payload
                                  FROM {self._table_name}
                                 WHERE TenantId = ?
                                   AND StreamId = ?
                                   AND StreamRevision <= ?
                                 ORDER BY CommitSequence;""",
                (commit.tenant_id, commit.stream_id, commit.stream_revision),
            )
            fetchall = await cur.fetchall()
            latest_revision = 0
            for doc in fetchall:
                events = [
                    EventMessage.from_dict(e, self._topic_map)
                    for e in from_json(doc[1])
                ]
                uncommitted_events = list(map(self._get_body, commit.events))
                committed_events = list(map(self._get_body, events))
                if self._conflict_detector.conflicts_with(
                    uncommitted_events, committed_events
                ):
                    return True, -1
                if doc[0] > latest_revision:
                    latest_revision = int(doc[0])
            return False, latest_revision

    @staticmethod
    def _get_body(e: EventMessage) -> BaseEvent:
        return e.body



import datetime
from typing import Iterable

from pydantic_core import from_json, to_json

from aett.domain import (
    ConflictDetector,
    DuplicateCommitException,
    ConflictingCommitException,
    NonConflictingCommitException,
)
from aett.eventstore import (
    ICommitEvents,
    TopicMap,
    COMMITS,
    MAX_INT,
    Commit,
    EventMessage,
    BaseEvent,
)
from aett.storage.synchronous.s3 import S3Config


class CommitStore(ICommitEvents):
    def __init__(
        self,
        s3_config: S3Config,
        topic_map: TopicMap,
        conflict_detector: ConflictDetector | None = None,
        folder_name=COMMITS,
    ):
        self._s3_bucket: str = s3_config.bucket
        self._topic_map: TopicMap = topic_map
        self._resource = s3_config.to_client()
        self._conflict_detector: ConflictDetector = (
            conflict_detector or ConflictDetector.empty()
        )
        self._folder_name = folder_name

    def get(
        self,
        tenant_id: str,
        stream_id: str,
        min_revision: int = 0,
        max_revision: int = MAX_INT,
    ) -> Iterable[Commit]:
        max_revision = MAX_INT if max_revision >= MAX_INT else max_revision + 1
        min_revision = 0 if min_revision < 0 else min_revision
        response = self._resource.list_objects(
            Bucket=self._s3_bucket,
            Delimiter="/",
            Prefix=f"{self._folder_name}/{tenant_id}/{stream_id}/",
        )
        if "Contents" not in response:
            return
        keys = [
            key
            for key in map(lambda r: r.get("Key"), response.get("Contents"))
            if min_revision
            <= int(key.split("_")[-1].replace(".json", ""))
            <= max_revision
        ]
        keys.sort()
        for key in keys:
            yield self._file_to_commit(key)

    def get_to(
        self,
        tenant_id: str,
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> Iterable[Commit]:
        response = self._resource.list_objects(
            Bucket=self._s3_bucket,
            Delimiter="/",
            Prefix=f"{self._folder_name}/{tenant_id}/{stream_id}/",
        )
        if "Contents" not in response:
            return
        timestamp = max_time.timestamp()
        keys = [
            key
            for key in map(lambda r: r.get("Key"), response.get("Contents"))
            if int(key.split("/")[-1].split("_")[0]) <= timestamp
        ]
        keys.sort()
        for key in keys:
            yield self._file_to_commit(key)

    def get_all_to(
        self, tenant_id: str, max_time: datetime.datetime = datetime.datetime.max
    ) -> Iterable[Commit]:
        response = self._resource.list_objects(
            Bucket=self._s3_bucket,
            Delimiter="/",
            Prefix=f"{self._folder_name}/{tenant_id}/",
        )
        if "Contents" not in response:
            return
        timestamp = max_time.timestamp()
        keys = [
            key
            for key in map(lambda r: r.get("Key"), response.get("Contents"))
            if int(key.split("/")[-1].split("_")[0]) <= timestamp
        ]
        keys.sort()
        for key in keys:
            yield self._file_to_commit(key)

    def _file_to_commit(self, key: str):
        file = self._resource.get_object(Bucket=self._s3_bucket, Key=key)
        doc = from_json(file.get("Body").read().decode("utf-8"))
        return Commit(
            tenant_id=doc.get("tenant_id"),
            stream_id=doc.get("stream_id"),
            stream_revision=doc.get("stream_revision"),
            commit_id=doc.get("commit_id"),
            commit_sequence=doc.get("commit_sequence"),
            commit_stamp=doc.get("commit_stamp"),
            headers=doc.get("headers"),
            events=[
                EventMessage.from_dict(e, self._topic_map) for e in doc.get("events")
            ],
            checkpoint_token=0,
        )

    def commit(self, commit: Commit):
        self.check_exists(commit_sequence=commit.commit_sequence, commit=commit)
        commit_key = f"{self._folder_name}/{commit.tenant_id}/{commit.stream_id}/{int(commit.commit_stamp.timestamp())}_{commit.commit_id}_{commit.commit_sequence}_{commit.stream_revision}.json"
        d = commit.__dict__
        d["events"] = [e.to_json() for e in commit.events]
        d["headers"] = {k: to_json(v) for k, v in commit.headers.items()}
        body = to_json(d)
        self._resource.put_object(
            Bucket=self._s3_bucket,
            Key=commit_key,
            Body=body,
            ContentLength=len(body),
            Metadata={k: to_json(v).decode("utf-8") for k, v in commit.headers.items()},
        )

    def check_exists(self, commit_sequence: int, commit: Commit):
        response = self._resource.list_objects(
            Delimiter="/",
            Prefix=f"{self._folder_name}/{commit.tenant_id}/{commit.stream_id}/",
            Bucket=self._s3_bucket,
        )
        if "Contents" not in response:
            return
        keys = list(
            key for key in map(lambda r: r.get("Key"), response.get("Contents"))
        )
        keys.sort()
        for key in keys:
            split = key.split("_")
            if commit.commit_id == split[1]:
                raise DuplicateCommitException(
                    f"Commit {commit.commit_id} already exists"
                )
            if int(split[-2]) == commit_sequence or commit.stream_revision <= int(
                split[-1].replace(".json", "")
            ):
                overlapping = [
                    key
                    for key in keys
                    if int(key.split("_")[-1].replace(".json", ""))
                    >= commit.stream_revision
                ]
                if len(overlapping) == 0:
                    return
                events = list(map(self._get_body, commit.events))
                for o in overlapping:
                    c = self._file_to_commit(o)
                    if self._conflict_detector.conflicts_with(
                        events, list(map(self._get_body, c.events))
                    ):
                        raise ConflictingCommitException(
                            f"Commit {commit.commit_id} conflicts with {c.commit_id}"
                        )
                raise NonConflictingCommitException(
                    f"Found non-conflicting commits at revision {commit.stream_revision}"
                )

    @staticmethod
    def _get_body(em: EventMessage) -> BaseEvent:
        body: BaseEvent = em.body
        return body



from typing import Iterable

import psycopg

from aett.eventstore import IManagePersistence, TopicMap, COMMITS, SNAPSHOTS, Commit
from aett.storage.synchronous.postgresql import _item_to_commit


class PersistenceManagement(IManagePersistence):
    def __init__(
        self,
        connection_string: str,
        topic_map: TopicMap,
        commits_table_name: str = COMMITS,
        snapshots_table_name: str = SNAPSHOTS,
    ):
        self._connection_string: str = connection_string
        self._topic_map = topic_map
        self._commits_table_name = commits_table_name
        self._snapshots_table_name = snapshots_table_name

    def initialize(self):
        try:
            with psycopg.connect(
                self._connection_string, autocommit=True
            ) as connection:
                with connection.cursor() as c:
                    c.execute(f"""CREATE TABLE {self._commits_table_name}
        (
            TenantId varchar(64) NOT NULL,
            StreamId char(64) NOT NULL,
            StreamIdOriginal varchar(1000) NOT NULL,
            StreamRevision int NOT NULL CHECK (StreamRevision > 0),
            Items smallint NOT NULL CHECK (Items > 0),
            CommitId uuid NOT NULL,
            CommitSequence int NOT NULL CHECK (CommitSequence > 0),
            CommitStamp timestamp NOT NULL,
            CheckpointNumber BIGSERIAL NOT NULL,
            Headers bytea NULL,
            Payload bytea NOT NULL,
            CONSTRAINT PK_Commits PRIMARY KEY (CheckpointNumber)
        );
        CREATE UNIQUE INDEX IX_Commits_CommitSequence ON {self._commits_table_name} (TenantId, StreamId, CommitSequence);
        CREATE UNIQUE INDEX IX_Commits_CommitId ON {self._commits_table_name} (TenantId, StreamId, CommitId);
        CREATE UNIQUE INDEX IX_Commits_Revisions ON {self._commits_table_name} (TenantId, StreamId, StreamRevision, Items);
        CREATE INDEX IX_Commits_Stamp ON {self._commits_table_name} (CommitStamp);
        
        CREATE TABLE {self._snapshots_table_name}
        (
            TenantId varchar(40) NOT NULL,
            StreamId char(40) NOT NULL,
            StreamRevision int NOT NULL CHECK (StreamRevision > 0),
            CommitSequence int NOT NULL CHECK (CommitSequence > 0),
            Payload bytea NOT NULL,
            Headers bytea NOT NULL,
            CONSTRAINT PK_Snapshots PRIMARY KEY (TenantId, StreamId, StreamRevision)
        );""")
                    c.commit()
        except Exception:
            pass

    def drop(self):
        with psycopg.connect(self._connection_string, autocommit=True) as connection:
            with connection.cursor() as c:
                c.execute(
                    f"""DROP TABLE {self._snapshots_table_name};DROP TABLE {self._commits_table_name};"""
                )

    def purge(self, tenant_id: str):
        with psycopg.connect(self._connection_string, autocommit=True) as connection:
            with connection.cursor() as c:
                c.execute(
                    f"""DELETE FROM {self._commits_table_name} WHERE TenantId = %s;""",
                    tenant_id,
                )
                c.execute(
                    f"""DELETE FROM {self._snapshots_table_name} WHERE TenantId = %s;""",
                    tenant_id,
                )

    def get_from(self, checkpoint: int) -> Iterable[Commit]:
        with psycopg.connect(self._connection_string, autocommit=True) as connection:
            with connection.cursor() as cur:
                cur.execute(
                    f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                                  FROM {self._commits_table_name}
                                 WHERE CommitStamp >= %s
                                 ORDER BY CheckpointNumber;""",
                    (checkpoint,),
                )
                fetchall = cur.fetchall()
                for doc in fetchall:
                    yield _item_to_commit(doc, self._topic_map)



from typing import TypeVar

from aett.eventstore import BaseEvent, Memento
from aett.eventstore.base_command import BaseCommand

TMemento = TypeVar("TMemento", bound=Memento, contravariant=True)
TCommand = TypeVar("TCommand", bound=BaseCommand, contravariant=True)
TEvent = TypeVar("TEvent", bound=BaseEvent, contravariant=True)

TUncommitted = TypeVar("TUncommitted", bound=BaseEvent)
TCommitted = TypeVar("TCommitted", bound=BaseEvent)

UNDISPATCHEDMESSAGES = "UndispatchedMessage"



from typing import Iterable

from pymongo import ASCENDING
from pymongo.cursor import Cursor
from pymongo.errors import CollectionInvalid
from pymongo import database

from aett.eventstore import IManagePersistence, TopicMap, COMMITS, SNAPSHOTS, Commit
from aett.storage.synchronous.mongodb import _doc_to_commit


class PersistenceManagement(IManagePersistence):
    def __init__(
        self,
        db: database.Database,
        topic_map: TopicMap,
        commits_table_name: str = COMMITS,
        snapshots_table_name: str = SNAPSHOTS,
    ):
        self._topic_map = topic_map
        self.db: database.Database = db
        self.commits_table_name = commits_table_name
        self.snapshots_table_name = snapshots_table_name

    def initialize(self):
        try:
            counters_collection: database.Collection = self.db.create_collection(
                "counters", check_exists=True
            )
            if counters_collection.count_documents({"_id": "CheckpointToken"}) == 0:
                counters_collection.insert_one({"_id": "CheckpointToken", "seq": 0})
        except CollectionInvalid:
            pass
        try:
            commits_collection: database.Collection = self.db.create_collection(
                self.commits_table_name, check_exists=True
            )
            commits_collection.create_index(
                [
                    ("TenantId", ASCENDING),
                    ("CheckpointToken", ASCENDING),
                ],
                comment="GetFromCheckpoint",
                unique=True,
            )
            commits_collection.create_index(
                [
                    ("TenantId", ASCENDING),
                    ("StreamId", ASCENDING),
                    ("StreamRevision", ASCENDING),
                ],
                comment="GetFrom",
                unique=True,
            )
            commits_collection.create_index(
                [
                    ("TenantId", ASCENDING),
                    ("StreamId", ASCENDING),
                    ("CommitSequence", ASCENDING),
                ],
                comment="LogicalKey",
                unique=True,
            )
            commits_collection.create_index(
                [("CommitStamp", ASCENDING)],
                comment="CommitStamp",
                unique=False,
            )
            commits_collection.create_index(
                [
                    ("TenantId", ASCENDING),
                    ("StreamId", ASCENDING),
                    ("CommitId", ASCENDING),
                ],
                comment="CommitId",
                unique=True,
            )
        except CollectionInvalid:
            pass

        try:
            snapshots_collection: database.Collection = self.db.create_collection(
                self.snapshots_table_name, check_exists=True
            )
            snapshots_collection.create_index(
                [
                    ("TenantId", ASCENDING),
                    ("StreamId", ASCENDING),
                    ("StreamRevision", ASCENDING),
                ],
                comment="LogicalKey",
                unique=True,
            )
        except CollectionInvalid:
            pass

    def drop(self):
        self.db.drop_collection(self.commits_table_name)
        self.db.drop_collection(self.snapshots_table_name)

    def purge(self, tenant_id: str):
        collection = self.db.get_collection(self.commits_table_name)
        collection.delete_many({"TenantId": tenant_id})

    def get_from(self, checkpoint: int) -> Iterable[Commit]:
        collection = self.db.get_collection(self.commits_table_name)
        filters = {"CommitSequence": {"$gte": checkpoint}}
        query_response: Cursor = collection.find({"$and": [filters]})
        for doc in query_response.sort("CheckpointToken", direction=ASCENDING):
            yield _doc_to_commit(doc, self._topic_map)



import datetime
import typing
from abc import ABC, abstractmethod
from typing import Iterable

from aett.eventstore.constants import MAX_INT
from aett.eventstore.commit import Commit


class ICommitEvents(ABC):
    """
    Indicates the ability to commit events and access events to and from a given stream.

    Instances of this class must be designed to be multi-thread safe such that they can be shared between threads.
    """

    @abstractmethod
    def get(
        self,
        tenant_id: str,
        stream_id: str,
        min_revision: int = 0,
        max_revision: int = MAX_INT,
    ) -> Iterable[Commit]:
        """
        Gets the corresponding commits from the stream indicated starting at the revision specified until the
        end of the stream sorted in ascending order--from oldest to newest.

        :param tenant_id: The value which uniquely identifies bucket the stream belongs to.
        :param stream_id: The stream from which the events will be read.
        :param min_revision: The minimum revision of the stream to be read.
        :param max_revision: The maximum revision of the stream to be read.
        :return: A series of committed events from the stream specified sorted in ascending order.
        :raises StorageException:
        :raises StorageUnavailableException:
        """
        pass

    @abstractmethod
    def get_to(
        self,
        tenant_id: str,
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> typing.Iterable[Commit]:
        """
        Gets the corresponding commits from the stream indicated starting at the revision specified until the
        end of the stream sorted in ascending order--from oldest to newest.

        :param tenant_id: The value which uniquely identifies bucket the stream belongs to.
        :param stream_id: The stream from which the events will be read.
        :param max_time: The max timestamp to return.
        :return: A series of committed events from the stream specified sorted in ascending order.
        :raises StorageException:
        :raises StorageUnavailableException:
        """
        pass

    @abstractmethod
    def get_all_to(
        self, tenant_id: str, max_time: datetime.datetime = datetime.datetime.max
    ) -> Iterable[Commit]:
        """
        Gets the corresponding commits from the stream indicated starting at the revision specified until the
        end of the stream sorted in ascending order--from oldest to newest.

        :param tenant_id: The value which uniquely identifies bucket the stream belongs to.
        :param max_time: The max timestamp to return.
        :return: A series of committed events from the stream specified sorted in ascending order.
        :raises StorageException:
        :raises StorageUnavailableException:
        """
        pass

    @abstractmethod
    def commit(self, commit: Commit):
        """
        Writes the to-be-committed events stream provided to the underlying persistence mechanism.

        :param commit: The series of events and associated metadata to be committed.
        :raises ConcurrencyException:
        :raises StorageException:
        :raises StorageUnavailableException:
        """
        pass



import inspect
import logging
from typing import List, Dict, Type, Callable, Iterable

from aett.domain.conflict_delegate import ConflictDelegate
from aett.eventstore import BaseEvent, DomainEvent


class ConflictDetector:
    @staticmethod
    def empty() -> "ConflictDetector":
        return ConflictDetector()

    def __init__(
        self,
        delegates: List[ConflictDelegate] | None = None,
        logger: logging.Logger | None = None,
    ):
        """
        Initialize the conflict detector with the specified delegates.

        :param delegates: The delegates to use for conflict detection
        :param logger: The optional logger to use for logging.
        """
        self.delegates: Dict[
            Type, Dict[Type, Callable[[BaseEvent, BaseEvent], bool]]
        ] = {}
        self._logger = (
            logger
            if logger is not None and delegates is not None
            else logging.getLogger(ConflictDetector.__name__)
        )
        if delegates is not None:
            for delegate in delegates:
                args = inspect.getfullargspec(delegate.detect)
                uncommitted_type = args.annotations[args.args[1]]
                committed_type = args.annotations[args.args[2]]
                if uncommitted_type not in self.delegates:
                    self.delegates[uncommitted_type] = {}
                self.delegates[uncommitted_type][committed_type] = delegate.detect

    def conflicts_with(
        self,
        uncommitted_events: Iterable[BaseEvent],
        committed_events: Iterable[BaseEvent],
    ) -> bool:
        """
        Detects if the uncommitted events conflict with the committed events.

        :param uncommitted_events: The uncommitted events to analyze
        :param committed_events: The committed events to compare against.
        """
        if len(self.delegates) == 0:
            return False
        for uncommitted in uncommitted_events:
            for committed in committed_events:
                uncommitted_type = type(uncommitted)
                delegates_keys = self.delegates.keys()
                committed_type = type(committed)
                if uncommitted_type in delegates_keys:
                    committed_keys = self.delegates[uncommitted_type].keys()
                    if committed_type in committed_keys:
                        if self.delegates[uncommitted_type][committed_type](
                            uncommitted, committed
                        ):
                            if isinstance(uncommitted, DomainEvent):
                                self._logger.warning(
                                    f"Detected conflict between uncommitted event {uncommitted_type.__name__} from {uncommitted.source} with version {uncommitted.version}"
                                )
                            else:
                                self._logger.warning(
                                    f"Detected conflict between uncommitted event {uncommitted_type.__name__} from {uncommitted.source} with timestamp {uncommitted.timestamp:%Y%m%d-%H%M%S%z}"
                                )
                            return True
        return False



from typing import AsyncIterable

import pymongo
from pymongo.asynchronous import collection, cursor
from pymongo.asynchronous.collection import AsyncCollection
from pymongo.asynchronous.database import AsyncDatabase
from pymongo.errors import CollectionInvalid

from aett.eventstore import (
    Commit,
    IManagePersistenceAsync,
    TopicMap,
    COMMITS,
    SNAPSHOTS,
)
from aett.storage.asynchronous.mongodb.mapping import _doc_to_commit


class AsyncPersistenceManagement(IManagePersistenceAsync):
    def __init__(
        self,
        db: AsyncDatabase,
        topic_map: TopicMap,
        commits_table_name: str = COMMITS,
        snapshots_table_name: str = SNAPSHOTS,
    ):
        self._topic_map = topic_map
        self.db: AsyncDatabase = db
        self.commits_table_name = commits_table_name
        self.snapshots_table_name = snapshots_table_name

    async def initialize(self):
        try:
            counters_collection: AsyncCollection = await self.db.create_collection(
                "counters", check_exists=True
            )
            if (
                await counters_collection.count_documents({"_id": "CheckpointToken"})
                == 0
            ):
                await counters_collection.insert_one(
                    {"_id": "CheckpointToken", "seq": 0}
                )
        except pymongo.errors.CollectionInvalid:
            pass
        try:
            commits_collection: collection.AsyncCollection = (
                await self.db.create_collection(
                    self.commits_table_name, check_exists=True
                )
            )
            await commits_collection.create_index(
                [
                    ("TenantId", pymongo.ASCENDING),
                    ("CheckpointToken", pymongo.ASCENDING),
                ],
                comment="GetFromCheckpoint",
                unique=True,
            )
            await commits_collection.create_index(
                [
                    ("TenantId", pymongo.ASCENDING),
                    ("StreamId", pymongo.ASCENDING),
                    ("StreamRevision", pymongo.ASCENDING),
                ],
                comment="GetFrom",
                unique=True,
            )
            await commits_collection.create_index(
                [
                    ("TenantId", pymongo.ASCENDING),
                    ("StreamId", pymongo.ASCENDING),
                    ("CommitSequence", pymongo.ASCENDING),
                ],
                comment="LogicalKey",
                unique=True,
            )
            await commits_collection.create_index(
                [("CommitStamp", pymongo.ASCENDING)],
                comment="CommitStamp",
                unique=False,
            )
            await commits_collection.create_index(
                [
                    ("TenantId", pymongo.ASCENDING),
                    ("StreamId", pymongo.ASCENDING),
                    ("CommitId", pymongo.ASCENDING),
                ],
                comment="CommitId",
                unique=True,
            )
        except CollectionInvalid:
            pass

        try:
            snapshots_collection: collection.AsyncCollection = (
                await self.db.create_collection(
                    self.snapshots_table_name, check_exists=True
                )
            )
            await snapshots_collection.create_index(
                [
                    ("TenantId", pymongo.ASCENDING),
                    ("StreamId", pymongo.ASCENDING),
                    ("StreamRevision", pymongo.ASCENDING),
                ],
                comment="LogicalKey",
                unique=True,
            )
        except pymongo.errors.CollectionInvalid:
            pass

    async def drop(self):
        await self.db.drop_collection(self.commits_table_name)
        await self.db.drop_collection(self.snapshots_table_name)

    async def purge(self, tenant_id: str):
        c = self.db.get_collection(self.commits_table_name)
        await c.delete_many({"TenantId": tenant_id})

    async def get_from(self, checkpoint: int) -> AsyncIterable[Commit]:
        c = self.db.get_collection(self.commits_table_name)
        filters = {"CommitSequence": {"$gte": checkpoint}}
        query_response: cursor.AsyncCursor = c.find({"$and": [filters]})
        async for doc in query_response.sort(
            "CheckpointToken", direction=pymongo.ASCENDING
        ):
            yield _doc_to_commit(doc, self._topic_map)



from typing import Dict

from pydantic_core import from_json, to_json

from aett.eventstore import IAccessSnapshotsAsync, SNAPSHOTS, Snapshot, MAX_INT
from aett.storage.asynchronous.s3 import AsyncS3Config


class AsyncSnapshotStore(IAccessSnapshotsAsync):
    def __init__(self, s3_config: AsyncS3Config, folder_name: str = SNAPSHOTS):
        self.__s3_bucket = s3_config.bucket
        self.__folder_name = folder_name
        self.__config = s3_config

    async def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        async with self.__config.to_client() as client:
            files = await client.list_objects(
                Bucket=self.__s3_bucket,
                Delimiter="/",
                Prefix=f"{self.__folder_name}/{tenant_id}/{stream_id}/",
            )
            if "Contents" not in files:
                return None
            keys = list(
                int(key.split("/")[-1].replace(".json", ""))
                for key in map(lambda r: r.get("Key"), files.get("Contents"))
                if int(key.split("/")[-1].replace(".json", "")) <= max_revision
            )
            keys.sort(reverse=True)

            key = f"{self.__folder_name}/{tenant_id}/{stream_id}/{keys[0]}.json"
            j = await client.get_object(Bucket=self.__s3_bucket, Key=key)
            body = await j["Body"].read()
            d = from_json(data=body)
            return Snapshot(
                tenant_id=d.get("tenant_id"),
                stream_id=d.get("stream_id"),
                stream_revision=int(d.get("stream_revision")),
                commit_sequence=int(d.get("commit_sequence")),
                payload=d.get("payload"),
                headers=d.get("headers"),
            )

    async def add(
        self, snapshot: Snapshot, headers: Dict[str, str] | None = None
    ) -> None:
        async with self.__config.to_client() as client:
            if headers is not None:
                snapshot.headers.update(headers)
            key = f"{self.__folder_name}/{snapshot.tenant_id}/{snapshot.stream_id}/{snapshot.stream_revision}.json"
            await client.put_object(
                Bucket=self.__s3_bucket, Key=key, Body=to_json(snapshot)
            )



import typing

from aett.eventstore import IAccessSnapshots, Snapshot, MAX_INT


class SnapshotStore(IAccessSnapshots):
    def __init__(self):
        self.buckets: typing.Dict[str, typing.Dict[str, typing.List[Snapshot]]] = {}

    def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        if not self._ensure_stream(tenant_id=tenant_id, stream_id=stream_id):
            return None
        if len(self.buckets[tenant_id][stream_id]) == 0:
            return None
        snapshots = list(
            filter(
                lambda s: s.stream_revision <= max_revision,
                self.buckets[tenant_id][stream_id],
            )
        )
        return snapshots[-1]

    def add(self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None):
        self._ensure_stream(snapshot.tenant_id, snapshot.stream_id)
        stream = self.buckets[snapshot.tenant_id][snapshot.stream_id]
        if len(stream) == 0:
            stream.append(snapshot)
        else:
            latest = stream[-1]
            if latest.stream_revision <= snapshot.stream_revision:
                raise ValueError("Conflicting commit")
            stream.append(snapshot)
            stream.sort(key=lambda s: s.stream_revision, reverse=False)

    def _ensure_stream(self, tenant_id: str, stream_id: str) -> bool:
        if tenant_id not in self.buckets:
            self.buckets[tenant_id] = {stream_id: []}
            return False
        if stream_id not in self.buckets[tenant_id]:
            self.buckets[tenant_id][stream_id] = list()
            return False
        return True



import logging

import aiosqlite
from typing import Iterable, AsyncIterable

from aett.eventstore import (
    IManagePersistenceAsync,
    TopicMap,
    COMMITS,
    SNAPSHOTS,
    Commit,
)
from aett.storage.asynchronous.sqlite.functions import _item_to_commit


class AsyncPersistenceManagement(IManagePersistenceAsync):
    def __init__(
        self,
        connection_string: str,
        topic_map: TopicMap,
        commits_table_name: str = COMMITS,
        snapshots_table_name: str = SNAPSHOTS,
    ):
        self._connection_string: str = connection_string
        self._topic_map = topic_map
        self._commits_table_name = commits_table_name
        self._snapshots_table_name = snapshots_table_name

    async def initialize(self):
        try:
            async with aiosqlite.connect(self._connection_string) as connection:
                c: aiosqlite.Cursor = await connection.cursor()
                await c.execute(f"""CREATE TABLE IF NOT EXISTS {self._commits_table_name}
    (
           TenantId varchar(40) NOT NULL,
           StreamId char(40) NOT NULL,
           StreamIdOriginal text NOT NULL,
           StreamRevision int NOT NULL CHECK (StreamRevision > 0),
           Items int NOT NULL CHECK (Items > 0),
           CommitId guid NOT NULL CHECK (CommitId != 0),
           CommitSequence int NOT NULL CHECK (CommitSequence > 0),
           CommitStamp datetime NOT NULL,
           CheckpointNumber INTEGER PRIMARY KEY AUTOINCREMENT,
           Headers blob NULL,
           Payload blob NOT NULL
    );""")
                await c.execute(
                    f"CREATE UNIQUE INDEX IF NOT EXISTS IX_Commits_CommitSequence ON {self._commits_table_name} (TenantId, StreamId, CommitSequence);"
                )
                await c.execute(
                    f"CREATE UNIQUE INDEX IF NOT EXISTS IX_Commits_CommitId ON {self._commits_table_name} (TenantId, StreamId, CommitId);"
                )
                await c.execute(
                    f"CREATE UNIQUE INDEX IF NOT EXISTS IX_Commits_Revisions ON {self._commits_table_name} (TenantId, StreamId, StreamRevision, Items);"
                )
                await c.execute(
                    f"CREATE INDEX IF NOT EXISTS IX_Commits_Stamp ON {self._commits_table_name} (CommitStamp);"
                )
                await c.execute(f"""CREATE TABLE IF NOT EXISTS {self._snapshots_table_name}
    (
           TenantId varchar(40) NOT NULL,
           StreamId char(40) NOT NULL,
           StreamRevision int NOT NULL CHECK (StreamRevision > 0),
           CommitSequence int NOT NULL CHECK (CommitSequence > 0),
           Payload blob NOT NULL,
           CONSTRAINT PK_Snapshots PRIMARY KEY (TenantId, StreamId, StreamRevision)
    );""")
                await connection.commit()
        except Exception as e:
            logging.error(
                f"Failed to initialize persistence with error {e}", exc_info=True
            )
            pass

    async def drop(self):
        async with aiosqlite.connect(self._connection_string, timeout=30) as connection:
            c: aiosqlite.Cursor = await connection.cursor()
            await c.execute(f"DROP TABLE {self._snapshots_table_name};")
            await c.execute(f"DROP TABLE {self._commits_table_name};")
            await connection.commit()

    async def purge(self, tenant_id: str):
        async with aiosqlite.connect(self._connection_string) as connection:
            c: aiosqlite.Cursor = await connection.cursor()
            await c.execute(
                f"""DELETE FROM {self._commits_table_name} WHERE TenantId = %s;""",
                tenant_id,
            )
            await c.execute(
                f"""DELETE FROM {self._snapshots_table_name} WHERE TenantId = %s;""",
                tenant_id,
            )
            await connection.commit()

    async def get_from(self, checkpoint: int) -> AsyncIterable[Commit]:
        async with aiosqlite.connect(self._connection_string) as connection:
            cur: aiosqlite.Cursor = await connection.cursor()
            await cur.execute(
                f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                                  FROM {self._commits_table_name}
                                 WHERE CommitStamp >= %s
                                 ORDER BY CheckpointNumber;""",
                (checkpoint,),
            )
            fetchall = await cur.fetchall()
            for doc in fetchall:
                yield _item_to_commit(doc, self._topic_map)



import typing

import psycopg
from pydantic_core import from_json, to_json

from aett.eventstore import IAccessSnapshots, SNAPSHOTS, MAX_INT, Snapshot


class SnapshotStore(IAccessSnapshots):
    def __init__(self, connection_string: str, table_name: str = SNAPSHOTS):
        self._connection_string: str = connection_string
        self._table_name = table_name

    def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        try:
            with psycopg.connect(
                self._connection_string, autocommit=True
            ) as connection:
                with connection.cursor() as cur:
                    cur.execute(
                        f"""SELECT *
          FROM {self._table_name}
         WHERE TenantId = %s
           AND StreamId = %s
           AND StreamRevision <= %s
         ORDER BY StreamRevision DESC
         LIMIT 1;""",
                        (tenant_id, stream_id, max_revision),
                    )
                    item = cur.fetchone()
                    if item is None:
                        return None

                    return Snapshot(
                        tenant_id=item[0],
                        stream_id=item[1],
                        stream_revision=int(item[2]),
                        commit_sequence=int(item[3]),
                        payload=from_json(item[4]),
                        headers=dict(from_json(item[5])),
                    )
        except Exception as e:
            raise Exception(
                f"Failed to get snapshot for stream {stream_id} with error {e}"
            )

    def add(self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None):
        if headers is None:
            headers = {}
        try:
            with psycopg.connect(
                self._connection_string, autocommit=True
            ) as connection:
                with connection.cursor() as cur:
                    cur.execute(
                        f"""INSERT INTO {self._table_name} ( TenantId, StreamId, StreamRevision, CommitSequence, Payload, Headers) VALUES (%s, %s, %s, %s, %s, %s);""",
                        (
                            snapshot.tenant_id,
                            snapshot.stream_id,
                            snapshot.stream_revision,
                            snapshot.commit_sequence,
                            to_json(snapshot.payload),
                            to_json(headers),
                        ),
                    )
                    connection.commit()
        except Exception as e:
            raise Exception(
                f"Failed to add snapshot for stream {snapshot.stream_id} with error {e}"
            )



from aett.storage.synchronous.sqlite.functions import _item_to_commit
from aett.storage.synchronous.sqlite.commit_store import CommitStore
from aett.storage.synchronous.sqlite.snapshot_store import SnapshotStore
from aett.storage.synchronous.sqlite.persistence_management import PersistenceManagement



import datetime
from uuid import UUID

from pydantic_core import from_json

from aett.eventstore import TopicMap, Commit, EventMessage


def _doc_to_commit(doc: dict, topic_map: TopicMap) -> Commit:
    loads = from_json(doc["Events"])
    events_ = [EventMessage.from_dict(e, topic_map) for e in loads]
    return Commit(
        tenant_id=doc["TenantId"],
        stream_id=doc["StreamId"],
        stream_revision=int(doc["StreamRevision"]),
        commit_id=UUID(doc["CommitId"]),
        commit_sequence=int(doc["CommitSequence"]),
        commit_stamp=datetime.datetime.fromtimestamp(
            int(doc["CommitStamp"]), datetime.timezone.utc
        ),
        headers=from_json(doc["Headers"]),
        events=events_,
        checkpoint_token=doc["CheckpointToken"],
    )



from typing import Dict, Any

from pydantic import BaseModel, Field

from aett.eventstore.topic import Topic
from aett.eventstore.topic_map import TopicMap


TOPIC_HEADER = "topic"


class EventMessage(BaseModel):
    """
    Represents a single event message within a commit.
    """

    body: Any = Field(description="Gets the body of the event message.")

    headers: Dict[str, Any] | None = Field(
        default_factory=dict,
        description="Gets the metadata which provides additional, "
        "unstructured information about this event message.",
    )

    def to_json(self) -> dict:
        """
        Converts the event message to a dictionary which can be serialized to JSON.
        """
        if self.headers is None:
            self.headers = {}
        if "topic" not in self.headers:
            self.headers[TOPIC_HEADER] = Topic.get(type(self.body))
        return self.model_dump(serialize_as_any=True)

    @staticmethod
    def from_dict(json_dict: dict, topic_map: TopicMap) -> "EventMessage":
        headers = (
            json_dict["headers"]
            if "headers" in json_dict and json_dict["headers"] is not None
            else {}
        )
        decoded_body = json_dict["body"]
        topic = (
            decoded_body.pop("$type", None) if isinstance(decoded_body, dict) else None
        )
        if topic is None and TOPIC_HEADER in headers:
            topic = headers[TOPIC_HEADER]
        if topic is None:
            return EventMessage(body=decoded_body, headers=headers)
        else:
            t = topic_map.get(topic=topic)
            body = (
                t.model_validate(decoded_body)
                if t is not None and issubclass(t, BaseModel)
                else decoded_body
            )
            return EventMessage(body=body, headers=headers)



from aett.storage.asynchronous.s3.async_s3_config import AsyncS3Config
from aett.storage.asynchronous.s3.async_persistence_management import (
    AsyncPersistenceManagement,
)
from aett.storage.asynchronous.s3.async_snapshot_store import AsyncSnapshotStore
from aett.storage.asynchronous.s3.async_commit_store import AsyncCommitStore



from typing import Iterable

from aioboto3 import Session
from boto3.dynamodb.conditions import Key

from aett.eventstore import Commit
from aett.eventstore.constants import COMMITS, SNAPSHOTS
from aett.eventstore.i_manage_persistence_async import IManagePersistenceAsync
from aett.storage.asynchronous.dynamodb import _get_client


class AsyncPersistenceManagement(IManagePersistenceAsync):
    def __init__(
        self,
        region: str = "eu-central-1",
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        port: int = 8000,
        commits_table_name: str = COMMITS,
        snapshots_table_name: str = SNAPSHOTS,
    ):
        self.__session = Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region,
            profile_name=profile_name,
        )
        self.__port = port
        self.commits_table_name = commits_table_name
        self.snapshots_table_name = snapshots_table_name

    async def initialize(self):
        async with _get_client(session=self.__session, port=self.__port) as client:
            tables = client.tables.all()
            table_names = [table.name async for table in tables]
            if self.commits_table_name not in table_names:
                _ = await client.create_table(
                    TableName=self.commits_table_name,
                    KeySchema=[
                        {"AttributeName": "TenantAndStream", "KeyType": "HASH"},
                        {"AttributeName": "CommitSequence", "KeyType": "RANGE"},
                    ],
                    AttributeDefinitions=[
                        {"AttributeName": "TenantAndStream", "AttributeType": "S"},
                        {"AttributeName": "CommitSequence", "AttributeType": "N"},
                        {"AttributeName": "StreamRevision", "AttributeType": "N"},
                        {"AttributeName": "CommitStamp", "AttributeType": "N"},
                    ],
                    LocalSecondaryIndexes=[
                        {
                            "IndexName": "RevisionIndex",
                            "KeySchema": [
                                {"AttributeName": "TenantAndStream", "KeyType": "HASH"},
                                {"AttributeName": "StreamRevision", "KeyType": "RANGE"},
                            ],
                            "Projection": {"ProjectionType": "ALL"},
                        },
                        {
                            "IndexName": "CommitStampIndex",
                            "KeySchema": [
                                {"AttributeName": "TenantAndStream", "KeyType": "HASH"},
                                {"AttributeName": "CommitStamp", "KeyType": "RANGE"},
                            ],
                            "Projection": {"ProjectionType": "ALL"},
                        },
                    ],
                    TableClass="STANDARD",
                    StreamSpecification={
                        "StreamEnabled": True,
                        "StreamViewType": "NEW_IMAGE",
                    },
                    ProvisionedThroughput={
                        "ReadCapacityUnits": 10,
                        "WriteCapacityUnits": 10,
                    },
                )

            if self.snapshots_table_name not in table_names:
                _ = await client.create_table(
                    TableName=self.snapshots_table_name,
                    KeySchema=[
                        {"AttributeName": "TenantAndStream", "KeyType": "HASH"},
                        {"AttributeName": "StreamRevision", "KeyType": "RANGE"},
                    ],
                    AttributeDefinitions=[
                        {"AttributeName": "TenantAndStream", "AttributeType": "S"},
                        {"AttributeName": "StreamRevision", "AttributeType": "N"},
                    ],
                    TableClass="STANDARD",
                    ProvisionedThroughput={
                        "ReadCapacityUnits": 10,
                        "WriteCapacityUnits": 10,
                    },
                )

    async def drop(self):
        with _get_client(
            session=self.__session, region=self.__region, port=self.__port
        ) as client:
            tables = await client.tables.all()
            for table in tables:
                if table.name in [self.commits_table_name, self.snapshots_table_name]:
                    await table.delete()

    async def purge(self, tenant_id: str):
        with _get_client(
            session=self.__session, region=self.__region, port=self.__port
        ) as client:
            table = client.Table(self.commits_table_name)
            query_response = await table.scan(
                IndexName="CommitStampIndex",
                ConsistentRead=True,
                Select="ALL_ATTRIBUTES",
                ProjectionExpression="Tenant,CommitSequence",
                FilterExpression=(Key("Tenant").eq(f"{tenant_id}")),
            )
            with table.batch_writer() as batch:
                for each in query_response["Items"]:
                    await batch.delete_item(
                        Key={
                            "Tenant": each["Tenant"],
                            "CommitSequence": each["CommitSequence"],
                        }
                    )

    async def get_from(self, checkpoint: int) -> Iterable[Commit]:
        raise NotImplementedError()



from abc import ABC, abstractmethod
from typing import Generic, List

from aett.domain.constants import TMemento
from aett.eventstore import DomainEvent, EventMessage


class Aggregate(ABC, Generic[TMemento]):
    """
    An aggregate is a cluster of domain objects that can be treated as a single unit. The aggregate base class requires
    implementors to provide a method to apply a snapshot and to get a memento.

    In addition to this, the aggregate base class provides a method to raise events, but the concrete application
    of the event relies on multiple dispatch to call the correct apply method in the subclass.
    """

    def __init__(
        self, stream_id: str, commit_sequence: int, memento: TMemento | None = None
    ):
        """
        Initialize the aggregate

        :param stream_id: The id of the stream
        :param commit_sequence: The commit sequence number which the aggregate was built from
        """
        self.uncommitted: List[EventMessage] = []
        self._id = stream_id
        self._version = 0
        self._commit_sequence = commit_sequence
        if memento is not None:
            self._version = memento.version
            self.apply_memento(memento)

    @property
    def id(self) -> str:
        """
        Gets the id of the aggregate
        """
        return self._id

    @property
    def version(self) -> int:
        """
        Gets the version of the aggregate
        """
        return self._version

    @property
    def commit_sequence(self):
        """
        Gets the commit sequence number of the aggregate
        """
        return self._commit_sequence

    @abstractmethod
    def apply_memento(self, memento: TMemento) -> None:
        """
        Apply a memento to the aggregate
        :param memento: The memento to apply
        :return: None
        """
        pass

    @abstractmethod
    def get_memento(self) -> TMemento:
        """
        Get a memento of the current state of the aggregate
        :return: A memento instance
        """
        pass

    def __getstate__(self):
        return self.get_memento()

    def __setstate__(self, state):
        self.apply_memento(state)

    def raise_event(self, event: DomainEvent) -> None:
        """
        Raise an event on the aggregate. This is the method that internal logic should use to raise events in order to
        ensure that the event gets applied and the version gets incremented and the event is made available for
        persistence in the event store.
        :param event:
        :return:
        """
        # Use multiple dispatch to call the correct apply method
        self._apply(event)
        self._version += 1
        self.uncommitted.append(EventMessage(body=event, headers=None))



import typing

import pymongo
from pydantic_core import from_json, to_json
from pymongo.asynchronous import database, collection

from aett.eventstore import IAccessSnapshotsAsync, SNAPSHOTS, MAX_INT, Snapshot


class AsyncSnapshotStore(IAccessSnapshotsAsync):
    def __init__(self, db: database.AsyncDatabase, table_name: str = SNAPSHOTS):
        self.collection: collection.AsyncCollection = db.get_collection(table_name)

    async def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        try:
            filters = {
                "TenantId": tenant_id,
                "StreamId": stream_id,
                "StreamRevision": {"$lte": max_revision},
            }
            cursor = (
                self.collection.find({"$and": [filters]})
                .sort("StreamRevision", direction=pymongo.DESCENDING)
                .limit(1)
            )
            items = await cursor.to_list(length=1)
            if not items:
                return None
            item = items[0]

            return Snapshot(
                tenant_id=item["TenantId"],
                stream_id=item["StreamId"],
                stream_revision=int(item["StreamRevision"]),
                commit_sequence=int(item["CommitSequence"]),
                payload=from_json(item["Payload"]),
                headers=from_json(item["Headers"]),
            )
        except Exception as e:
            raise Exception(
                f"Failed to get snapshot for stream {stream_id} with status code {e}"
            )

    async def add(
        self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None
    ):
        if headers is None:
            headers = {}
        try:
            doc = {
                "TenantId": snapshot.tenant_id,
                "StreamId": snapshot.stream_id,
                "StreamRevision": snapshot.stream_revision,
                "CommitSequence": snapshot.commit_sequence,
                "Payload": to_json(snapshot.payload),
                "Headers": to_json(headers),
            }
            _ = await self.collection.insert_one(doc)
        except Exception as e:
            raise Exception(
                f"Failed to add snapshot for stream {snapshot.stream_id} with status code {e}"
            )



from typing import Iterable

from boto3.dynamodb.conditions import Key

from aett.storage.synchronous.dynamodb import _get_resource
from aett.eventstore import Commit
from aett.eventstore.i_manage_persistence import IManagePersistence

from aett.eventstore.constants import COMMITS, SNAPSHOTS


class PersistenceManagement(IManagePersistence):
    def __init__(
        self,
        commits_table_name: str = COMMITS,
        snapshots_table_name: str = SNAPSHOTS,
        region: str = "eu-central-1",
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        port: int = 8000,
    ):
        self.dynamodb = _get_resource(
            profile_name=profile_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region=region,
            port=port,
        )
        self.commits_table_name = commits_table_name
        self.snapshots_table_name = snapshots_table_name

    def initialize(self):
        tables = self.dynamodb.tables.all()
        table_names = [table.name for table in tables]
        if self.commits_table_name not in table_names:
            response = self.dynamodb.create_table(
                TableName=self.commits_table_name,
                KeySchema=[
                    {"AttributeName": "TenantAndStream", "KeyType": "HASH"},
                    {"AttributeName": "CommitSequence", "KeyType": "RANGE"},
                ],
                AttributeDefinitions=[
                    {"AttributeName": "TenantAndStream", "AttributeType": "S"},
                    {"AttributeName": "CommitSequence", "AttributeType": "N"},
                    {"AttributeName": "StreamRevision", "AttributeType": "N"},
                    {"AttributeName": "CommitStamp", "AttributeType": "N"},
                ],
                LocalSecondaryIndexes=[
                    {
                        "IndexName": "RevisionIndex",
                        "KeySchema": [
                            {"AttributeName": "TenantAndStream", "KeyType": "HASH"},
                            {"AttributeName": "StreamRevision", "KeyType": "RANGE"},
                        ],
                        "Projection": {"ProjectionType": "ALL"},
                    },
                    {
                        "IndexName": "CommitStampIndex",
                        "KeySchema": [
                            {"AttributeName": "TenantAndStream", "KeyType": "HASH"},
                            {"AttributeName": "CommitStamp", "KeyType": "RANGE"},
                        ],
                        "Projection": {"ProjectionType": "ALL"},
                    },
                ],
                TableClass="STANDARD",
                StreamSpecification={
                    "StreamEnabled": True,
                    "StreamViewType": "NEW_IMAGE",
                },
                ProvisionedThroughput={
                    "ReadCapacityUnits": 10,
                    "WriteCapacityUnits": 10,
                },
            )

        if self.snapshots_table_name not in table_names:
            _ = self.dynamodb.create_table(
                TableName=self.snapshots_table_name,
                KeySchema=[
                    {"AttributeName": "TenantAndStream", "KeyType": "HASH"},
                    {"AttributeName": "StreamRevision", "KeyType": "RANGE"},
                ],
                AttributeDefinitions=[
                    {"AttributeName": "TenantAndStream", "AttributeType": "S"},
                    {"AttributeName": "StreamRevision", "AttributeType": "N"},
                ],
                TableClass="STANDARD",
                ProvisionedThroughput={
                    "ReadCapacityUnits": 10,
                    "WriteCapacityUnits": 10,
                },
            )

    def drop(self):
        tables = self.dynamodb.tables.all()
        for table in tables:
            if table.name in [self.commits_table_name, self.snapshots_table_name]:
                table.delete()

    def purge(self, tenant_id: str):
        table = self.dynamodb.Table(self.commits_table_name)
        query_response = table.scan(
            IndexName="CommitStampIndex",
            ConsistentRead=True,
            Select="ALL_ATTRIBUTES",
            ProjectionExpression="Tenant,CommitSequence",
            FilterExpression=(Key("Tenant").eq(f"{tenant_id}")),
        )
        with table.batch_writer() as batch:
            for each in query_response["Items"]:
                batch.delete_item(
                    Key={
                        "Tenant": each["Tenant"],
                        "CommitSequence": each["CommitSequence"],
                    }
                )

    def get_from(self, checkpoint: int) -> Iterable[Commit]:
        raise NotImplementedError()



import typing

from boto3.dynamodb.conditions import Key
from pydantic_core import from_json, to_json

from aett.storage.synchronous.dynamodb import _get_resource
from aett.eventstore import SNAPSHOTS, MAX_INT, Snapshot, IAccessSnapshots


class SnapshotStore(IAccessSnapshots):
    def __init__(
        self,
        table_name: str = SNAPSHOTS,
        region: str = "eu-central-1",
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        port: int = 8000,
    ):
        self.dynamodb = _get_resource(
            profile_name=profile_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region=region,
            port=port,
        )
        self.table = self.dynamodb.Table(table_name)
        self.table_name = table_name

    def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        try:
            query_response = self.table.query(
                TableName=self.table_name,
                ConsistentRead=True,
                Limit=1,
                KeyConditionExpression=(
                    Key("TenantAndStream").eq(f"{tenant_id}{stream_id}")
                    & Key("StreamRevision").lte(max_revision)
                ),
                ScanIndexForward=False,
            )
            if len(query_response["Items"]) == 0:
                return None
            item = query_response["Items"][0]
            return Snapshot(
                tenant_id=item["TenantId"],
                stream_id=item["StreamId"],
                stream_revision=int(item["StreamRevision"]),
                payload=item["Payload"],
                commit_sequence=item["CommitSequence"],
                headers=dict(from_json(item["Headers"])),
            )
        except Exception as e:
            raise Exception(
                f"Failed to get snapshot for stream {stream_id} with status code {e}"
            )

    def add(self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None):
        if headers is None:
            headers = {}
        try:
            item = {
                "TenantAndStream": f"{snapshot.tenant_id}{snapshot.stream_id}",
                "TenantId": snapshot.tenant_id,
                "StreamId": snapshot.stream_id,
                "StreamRevision": snapshot.stream_revision,
                "Payload": snapshot.payload,
                "CommitSequence": snapshot.commit_sequence,
                "Headers": to_json(headers).decode("utf-8"),
            }
            _ = self.table.put_item(
                TableName=self.table_name,
                Item=item,
                ReturnValues="NONE",
                ReturnValuesOnConditionCheckFailure="NONE",
                ConditionExpression="attribute_not_exists(TenantAndStream) AND attribute_not_exists(StreamRevision)",
            )
        except Exception as e:
            raise Exception(
                f"Failed to add snapshot for stream {snapshot.stream_id} with status code {e}"
            )



import datetime
import typing
from uuid import UUID

from asyncpg import connect
from asyncpg.exceptions import UniqueViolationError
from pydantic_core import to_json, from_json

from aett.domain import (
    ConflictDetector,
    DuplicateCommitException,
    ConflictingCommitException,
    NonConflictingCommitException,
)
from aett.eventstore import (
    ICommitEventsAsync,
    COMMITS,
    TopicMap,
    MAX_INT,
    Commit,
    EventMessage,
    BaseEvent,
)
from aett.storage.asynchronous.postgresql import _item_to_commit


class AsyncCommitStore(ICommitEventsAsync):
    def __init__(
        self,
        connection_string: str,
        topic_map: TopicMap,
        conflict_detector: ConflictDetector | None = None,
        table_name=COMMITS,
    ):
        self._topic_map = topic_map
        self._connection_string = connection_string
        self._conflict_detector = (
            conflict_detector if conflict_detector is not None else ConflictDetector()
        )
        self._table_name = table_name

    async def get(
        self,
        tenant_id: str,
        stream_id: str,
        min_revision: int = 0,
        max_revision: int = MAX_INT,
    ) -> typing.AsyncIterable[Commit]:
        max_revision = MAX_INT if max_revision >= MAX_INT else max_revision + 1
        min_revision = 0 if min_revision < 0 else min_revision
        connection = await connect(self._connection_string)
        fetchall = await connection.fetch(
            f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
          FROM {self._table_name}
         WHERE TenantId = $1
           AND StreamId = $2
           AND StreamRevision >= $3
           AND (StreamRevision - Items) < $4
           AND CommitSequence > $5
         ORDER BY CommitSequence;""",
            tenant_id,
            stream_id,
            min_revision,
            max_revision,
            0,
        )
        for doc in fetchall:
            yield _item_to_commit(doc, self._topic_map)

    async def get_to(
        self,
        tenant_id: str,
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> typing.AsyncIterable[Commit]:
        connection = await connect(self._connection_string)
        fetchall = await connection.fetch(
            f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                  FROM {self._table_name}
                 WHERE TenantId = $1
                   AND StreamId = $2
                   AND CommitStamp <= $3
                 ORDER BY CommitSequence;""",
            tenant_id,
            stream_id,
            max_time,
        )
        for doc in fetchall:
            yield _item_to_commit(doc, self._topic_map)

    async def get_all_to(
        self, tenant_id: str, max_time: datetime.datetime = datetime.datetime.max
    ) -> typing.AsyncIterable[Commit]:
        connection = await connect(self._connection_string)
        fetchall = await connection.fetch(
            f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                          FROM {self._table_name}
                         WHERE TenantId = %s
                           AND CommitStamp <= %s
                         ORDER BY CheckpointNumber;""",
            (tenant_id, max_time),
        )
        for doc in fetchall:
            yield _item_to_commit(doc, self._topic_map)

    async def commit(self, commit: Commit) -> Commit:
        try:
            connection = await connect(self._connection_string)
            json = to_json([e.to_json() for e in commit.events])
            fetchrow = await connection.fetchrow(
                f"""INSERT
              INTO {self._table_name}
                 ( TenantId, StreamId, StreamIdOriginal, CommitId, CommitSequence, StreamRevision, Items, CommitStamp, Headers, Payload )
            VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)
            RETURNING CheckpointNumber;""",
                commit.tenant_id,
                commit.stream_id,
                commit.stream_id,
                commit.commit_id,
                commit.commit_sequence,
                commit.stream_revision,
                len(commit.events),
                commit.commit_stamp,
                to_json(commit.headers),
                json,
            )
            checkpoint_number = fetchrow["checkpointnumber"]
            await connection.close()
            return Commit(
                tenant_id=commit.tenant_id,
                stream_id=commit.stream_id,
                stream_revision=commit.stream_revision,
                commit_id=commit.commit_id,
                commit_sequence=commit.commit_sequence,
                commit_stamp=commit.commit_stamp,
                headers=commit.headers,
                events=commit.events,
                checkpoint_token=checkpoint_number,
            )

        except UniqueViolationError:
            if await self._detect_duplicate(
                commit.commit_id, commit.tenant_id, commit.stream_id
            ):
                raise DuplicateCommitException(
                    f"Commit {commit.commit_id} already exists in stream {commit.stream_id}"
                )
            else:
                conflicts, revision = await self._detect_conflicts(commit=commit)
            if conflicts:
                raise ConflictingCommitException(
                    f"Conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                )
            else:
                raise NonConflictingCommitException(
                    f"Non-conflicting version conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                )
        except Exception as e:
            raise Exception(f"Failed to commit {commit.commit_id} with error {e}")

    async def _detect_duplicate(
        self, commit_id: UUID, tenant_id: str, stream_id: str
    ) -> bool:
        try:
            connection = await connect(self._connection_string)
            result = await connection.fetch(
                f"""SELECT COUNT(*)
                  FROM {self._table_name}
                 WHERE TenantId = $1
                   AND StreamId = $2
                   AND CommitId = $3;""",
                tenant_id,
                stream_id,
                str(commit_id),
            )
            await connection.close()
            count = result[0][0]
            return count > 0
        except Exception as e:
            raise Exception(
                f"Failed to detect duplicate commit {commit_id} with error {e}"
            )

    async def _detect_conflicts(self, commit: Commit) -> typing.Tuple[bool, int]:
        connection = await connect(self._connection_string)
        fetchall = await connection.fetch(
            f"""SELECT StreamRevision, Payload
                                  FROM {self._table_name}
                                 WHERE TenantId = $1
                                   AND StreamId = $2
                                   AND StreamRevision <= $3
                                 ORDER BY CommitSequence;""",
            commit.tenant_id,
            commit.stream_id,
            commit.stream_revision,
        )
        latest_revision = 0
        for doc in fetchall:
            events = [
                EventMessage.from_dict(e, self._topic_map) for e in from_json(doc[1])
            ]
            uncommitted_events = list(map(self._get_body, commit.events))
            committed_events = list(map(self._get_body, events))
            if self._conflict_detector.conflicts_with(
                uncommitted_events, committed_events
            ):
                return True, -1
            if doc[0] > latest_revision:
                latest_revision = int(doc[0])
        return False, latest_revision

    @staticmethod
    def _get_body(e: EventMessage) -> BaseEvent:
        return e.body



from typing import Iterable

from aett.eventstore import IManagePersistence, COMMITS, Commit, StreamHead
from aett.storage.synchronous.s3 import S3Config


class PersistenceManagement(IManagePersistence):
    def __init__(self, s3_config: S3Config, folder_name=COMMITS):
        self._folder_name = folder_name
        self._s3_bucket = s3_config.bucket
        self._resource = s3_config.to_client()

    def initialize(self):
        try:
            self._resource.create_bucket(Bucket=self._s3_bucket)
        except:
            pass

    def drop(self):
        self._resource.delete_bucket(Bucket=self._s3_bucket)

    def purge(self, tenant_id: str):
        response = self._resource.list_objects_v2(
            Bucket=self._s3_bucket, Prefix=f"{self._folder_name}/{tenant_id}/"
        )

        for o in response["Contents"]:
            self._resource.delete_object(Bucket=self._s3_bucket, Key=o["Key"])

    def get_from(self, checkpoint: int) -> Iterable[Commit]:
        return []

    def get_streams_to_snapshot(self, threshold: int) -> Iterable[StreamHead]:
        return []



import datetime
import typing
from typing import Iterable
from uuid import UUID

import psycopg
from pydantic_core import to_json, from_json

from aett.domain import (
    ConflictDetector,
    DuplicateCommitException,
    ConflictingCommitException,
    NonConflictingCommitException,
)
from aett.eventstore import (
    ICommitEvents,
    TopicMap,
    COMMITS,
    MAX_INT,
    Commit,
    EventMessage,
    BaseEvent,
)
from aett.storage.synchronous.postgresql import _item_to_commit


class CommitStore(ICommitEvents):
    def __init__(
        self,
        connection_string: str,
        topic_map: TopicMap | None = None,
        conflict_detector: ConflictDetector | None = None,
        table_name=COMMITS,
    ):
        self._topic_map = topic_map if topic_map else TopicMap()
        self._connection_string = connection_string
        self._conflict_detector = (
            conflict_detector if conflict_detector is not None else ConflictDetector()
        )
        self._table_name = table_name

    def get(
        self,
        tenant_id: str,
        stream_id: str,
        min_revision: int = 0,
        max_revision: int = MAX_INT,
    ) -> typing.Iterable[Commit]:
        max_revision = MAX_INT if max_revision >= MAX_INT else max_revision + 1
        min_revision = 0 if min_revision < 0 else min_revision
        with psycopg.connect(self._connection_string, autocommit=True) as connection:
            with connection.cursor() as cur:
                cur.execute(
                    f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
          FROM {self._table_name}
         WHERE TenantId = %s
           AND StreamId = %s
           AND StreamRevision >= %s
           AND (StreamRevision - Items) < %s
           AND CommitSequence > %s
         ORDER BY CommitSequence;""",
                    (tenant_id, stream_id, min_revision, max_revision, 0),
                )
                fetchall = cur.fetchall()
                for doc in fetchall:
                    yield _item_to_commit(doc, self._topic_map)

    def get_to(
        self,
        tenant_id: str,
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> Iterable[Commit]:
        with psycopg.connect(self._connection_string, autocommit=True) as connection:
            with connection.cursor() as cur:
                cur.execute(
                    f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                  FROM {self._table_name}
                 WHERE TenantId = %s
                   AND StreamId = %s
                   AND CommitStamp <= %s
                 ORDER BY CommitSequence;""",
                    (tenant_id, stream_id, max_time),
                )
                fetchall = cur.fetchall()
                for doc in fetchall:
                    yield _item_to_commit(doc, self._topic_map)

    def get_all_to(
        self, tenant_id: str, max_time: datetime.datetime = datetime.datetime.max
    ) -> Iterable[Commit]:
        with psycopg.connect(self._connection_string, autocommit=True) as connection:
            with connection.cursor() as cur:
                cur.execute(
                    f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                          FROM {self._table_name}
                         WHERE TenantId = %s
                           AND CommitStamp <= %s
                         ORDER BY CheckpointNumber;""",
                    (tenant_id, max_time),
                )
                fetchall = cur.fetchall()
                for doc in fetchall:
                    yield _item_to_commit(doc, self._topic_map)

    def commit(self, commit: Commit):
        try:
            with psycopg.connect(
                self._connection_string, autocommit=True
            ) as connection:
                with connection.cursor() as cur:
                    json = to_json([e.to_json() for e in commit.events])
                    cur.execute(
                        f"""INSERT
          INTO {self._table_name}
             ( TenantId, StreamId, StreamIdOriginal, CommitId, CommitSequence, StreamRevision, Items, CommitStamp, Headers, Payload )
        VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
        RETURNING CheckpointNumber;""",
                        (
                            commit.tenant_id,
                            commit.stream_id,
                            commit.stream_id,
                            commit.commit_id,
                            commit.commit_sequence,
                            commit.stream_revision,
                            len(commit.events),
                            commit.commit_stamp,
                            to_json(commit.headers),
                            json,
                        ),
                    )
                    checkpoint_number = cur.fetchone()
                    cur.close()
                    connection.commit()
                    return Commit(
                        tenant_id=commit.tenant_id,
                        stream_id=commit.stream_id,
                        stream_revision=commit.stream_revision,
                        commit_id=commit.commit_id,
                        commit_sequence=commit.commit_sequence,
                        commit_stamp=commit.commit_stamp,
                        headers=commit.headers,
                        events=commit.events,
                        checkpoint_token=checkpoint_number[0],
                    )
        except psycopg.errors.UniqueViolation:
            if self._detect_duplicate(
                commit.commit_id, commit.tenant_id, commit.stream_id
            ):
                raise DuplicateCommitException(
                    f"Commit {commit.commit_id} already exists in stream {commit.stream_id}"
                )
            else:
                conflicts, revision = self._detect_conflicts(commit=commit)
                if conflicts:
                    raise ConflictingCommitException(
                        f"Conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                    )
                else:
                    raise NonConflictingCommitException(
                        f"Non-conflicting version conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                    )
        except Exception as e:
            raise Exception(f"Failed to commit {commit.commit_id} with error {e}")

    def _detect_duplicate(
        self, commit_id: UUID, tenant_id: str, stream_id: str
    ) -> bool:
        try:
            with psycopg.connect(
                self._connection_string, autocommit=True
            ) as connection:
                with connection.cursor() as cur:
                    cur.execute(
                        f"""SELECT COUNT(*)
          FROM {self._table_name}
         WHERE TenantId = %s
           AND StreamId = %s
           AND CommitId = %s;""",
                        (tenant_id, stream_id, str(commit_id)),
                    )
                    result = cur.fetchone()
                    return result[0] > 0
        except Exception as e:
            raise Exception(
                f"Failed to detect duplicate commit {commit_id} with error {e}"
            )

    def _detect_conflicts(self, commit: Commit) -> typing.Tuple[bool, int]:
        with psycopg.connect(self._connection_string, autocommit=True) as connection:
            with connection.cursor() as cur:
                cur.execute(
                    f"""SELECT StreamRevision, Payload
                          FROM {self._table_name}
                         WHERE TenantId = %s
                           AND StreamId = %s
                           AND StreamRevision <= %s
                         ORDER BY CommitSequence;""",
                    (commit.tenant_id, commit.stream_id, commit.stream_revision),
                )
                fetchall = cur.fetchall()
                latest_revision = 0
                for doc in fetchall:
                    events = [
                        EventMessage.from_dict(e, self._topic_map)
                        for e in from_json(doc[1])
                    ]
                    uncommitted_events = list(map(self._get_body, commit.events))
                    committed_events = list(map(self._get_body, events))
                    if self._conflict_detector.conflicts_with(
                        uncommitted_events, committed_events
                    ):
                        return True, -1
                    if doc[0] > latest_revision:
                        latest_revision = int(doc[0])
                return False, latest_revision

    @staticmethod
    def _get_body(e: EventMessage) -> BaseEvent:
        return e.body



from aioboto3 import Session
from botocore.client import BaseClient


def _get_client(
    session: Session,
    port: int = 8000,
) -> BaseClient:
    return session.resource(
        "dynamodb",
        use_ssl=False if session.region_name == "localhost" else True,
        endpoint_url=f"http://localhost:{port}"
        if session.region_name == "localhost"
        else None,
    )



from aett.storage.synchronous.s3.s3_config import S3Config
from aett.storage.synchronous.s3.persistence_management import PersistenceManagement
from aett.storage.synchronous.s3.snapshot_store import SnapshotStore
from aett.storage.synchronous.s3.commit_store import CommitStore



import typing

from pymongo import DESCENDING
from pymongo.errors import PyMongoError
from pydantic_core import from_json, to_json
from pymongo import database

from aett.eventstore import Snapshot
from aett.eventstore.i_access_snapshots import IAccessSnapshots

from aett.eventstore.constants import SNAPSHOTS, MAX_INT


class SnapshotStore(IAccessSnapshots):
    def __init__(self, db: database.Database, table_name: str = SNAPSHOTS):
        self.collection: database.Collection = db.get_collection(table_name)

    def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        try:
            filters = {
                "TenantId": tenant_id,
                "StreamId": stream_id,
                "StreamRevision": {"$lte": max_revision},
            }
            cursor = (
                self.collection.find({"$and": [filters]})
                .sort("StreamRevision", direction=DESCENDING)
                .limit(1)
            )
            item = next(cursor, None)
            if item is None:
                return None

            return Snapshot(
                tenant_id=item["TenantId"],
                stream_id=item["StreamId"],
                stream_revision=int(item["StreamRevision"]),
                commit_sequence=int(item["CommitSequence"]),
                payload=from_json(item["Payload"]),
                headers=from_json(item["Headers"]),
            )
        except Exception as e:
            raise Exception(
                f"Failed to get snapshot for stream {stream_id} with status code {e}"
            )

    def add(self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None):
        if headers is None:
            headers = {}
        try:
            doc = {
                "TenantId": snapshot.tenant_id,
                "StreamId": snapshot.stream_id,
                "StreamRevision": snapshot.stream_revision,
                "CommitSequence": snapshot.commit_sequence,
                "Payload": to_json(snapshot.payload),
                "Headers": to_json(headers),
            }
            _ = self.collection.insert_one(doc)
        except Exception as e:
            raise Exception(
                f"Failed to add snapshot for stream {snapshot.stream_id} with status code {e}"
            )



from typing import Dict, Any

from pydantic import BaseModel

from aett.eventstore.memento import Memento


class Snapshot(BaseModel):
    """
    Represents a materialized view of a stream at specific revision.
    """

    tenant_id: str
    """
    Gets the value which uniquely identifies the tenant to which the stream belongs.
    """

    stream_id: str
    """
    Gets the value which uniquely identifies the stream to which the snapshot applies.
    """

    stream_revision: int
    """
    Gets the position at which the snapshot applies.
    """

    commit_sequence: int
    """
    Gets the commit sequence at which the snapshot applies.
    """

    payload: Any
    """
    Gets the snapshot or materialized view of the stream at the revision indicated.
    """

    headers: Dict[str, str]

    @staticmethod
    def from_memento(
        tenant_id: str, memento: Memento, commit_sequence: int, headers: Dict[str, str]
    ) -> "Snapshot":
        """
        Converts the memento to a snapshot which can be persisted.
        :param tenant_id: The value which uniquely identifies the bucket to which the stream belongs.
        :param memento:  The memento to be converted.
        :param commit_sequence: The commit sequence at which the snapshot applies.
        :param headers: The headers to assign to the snapshot
        :return:
        """
        return Snapshot(
            tenant_id=tenant_id,
            stream_id=memento.id,
            stream_revision=memento.version,
            payload=memento.payload.model_dump_json(serialize_as_any=True),
            headers=headers,
            commit_sequence=commit_sequence,
        )



import aiosqlite
import typing

from pydantic_core import from_json, to_json

from aett.eventstore import IAccessSnapshotsAsync, SNAPSHOTS, MAX_INT, Snapshot


class AsyncSnapshotStore(IAccessSnapshotsAsync):
    def __init__(self, connection_string: str, table_name: str = SNAPSHOTS):
        self._connection_string: str = connection_string
        self._table_name = table_name

    async def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        try:
            async with aiosqlite.connect(self._connection_string) as connection:
                cur = await connection.cursor()
                await cur.execute(
                    f"""SELECT *
              FROM {self._table_name}
              WHERE TenantId = ?
               AND StreamId = ?
               AND StreamRevision <= ?
             ORDER BY StreamRevision DESC
             LIMIT 1;""",
                    (tenant_id, stream_id, max_revision),
                )
                item = await cur.fetchone()
                if item is None:
                    return None

                return Snapshot(
                    tenant_id=item[0],
                    stream_id=item[1],
                    stream_revision=int(item[2]),
                    commit_sequence=int(item[3]),
                    payload=from_json(item[4]),
                    headers={},
                )

        except Exception as e:
            raise Exception(
                f"Failed to get snapshot for stream {stream_id} with error {e}"
            )

    async def add(
        self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None
    ):
        try:
            async with aiosqlite.connect(self._connection_string) as connection:
                cur = await connection.cursor()
                await cur.execute(
                    f"""INSERT INTO {self._table_name} ( TenantId, StreamId, StreamRevision, CommitSequence, Payload) VALUES (?, ?, ?, ?, ?);""",
                    (
                        snapshot.tenant_id,
                        snapshot.stream_id,
                        snapshot.stream_revision,
                        snapshot.commit_sequence,
                        to_json(snapshot.payload),
                    ),
                )
                await cur.close()
                await connection.commit()

        except Exception as e:
            raise Exception(
                f"Failed to add snapshot for stream {snapshot.stream_id} with error {e}"
            )



import datetime
import typing
from typing import Iterable

from aett.domain import (
    ConflictDetector,
    DuplicateCommitException,
    ConflictingCommitException,
    NonConflictingCommitException,
)
from aett.eventstore import ICommitEvents, Commit, MAX_INT, EventMessage, BaseEvent


class CommitStore(ICommitEvents):
    def __init__(self, conflict_detector: ConflictDetector | None = None):
        self._buckets: typing.Dict[str, typing.Dict[str, typing.List[Commit]]] = {}
        self._conflict_detector: ConflictDetector = (
            conflict_detector if conflict_detector is not None else ConflictDetector()
        )

    def get(
        self,
        tenant_id: str,
        stream_id: str,
        min_revision: int = 0,
        max_revision: int = MAX_INT,
    ) -> typing.Iterable[Commit]:
        if not self._ensure_stream(tenant_id=tenant_id, stream_id=stream_id):
            return []
        max_revision = MAX_INT if max_revision >= MAX_INT else max_revision
        min_revision = 0 if min_revision < 0 else min_revision
        commits: typing.List[Commit] = self._buckets[tenant_id][stream_id]
        return (
            commit
            for commit in commits
            if min_revision <= commit.stream_revision <= max_revision
        )

    def get_to(
        self,
        tenant_id: str,
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> Iterable[Commit]:
        if not self._ensure_stream(tenant_id=tenant_id, stream_id=stream_id):
            return []
        commits: typing.List[Commit] = self._buckets[tenant_id][stream_id]
        return (commit for commit in commits if commit.commit_stamp <= max_time)

    def get_all_to(
        self, tenant_id: str, max_time: datetime.datetime = datetime.datetime.max
    ) -> Iterable[Commit]:
        commits: typing.List[Commit] = []
        for bucket in self._buckets:
            for stream in self._buckets[bucket]:
                commits.extend(self.get_to(bucket, stream, max_time))
        commits.sort(key=lambda c: c.commit_stamp)
        return commits

    def commit(self, commit: Commit):
        self._ensure_stream(commit.tenant_id, commit.stream_id)
        existing = self._buckets[commit.tenant_id][commit.stream_id]
        if len(existing) > 0 and existing[-1].commit_sequence >= commit.commit_sequence:
            if existing[-1].commit_id == commit.commit_id:
                raise DuplicateCommitException("Duplicate commit")
            commits = [
                e
                for c in (
                    c.events
                    for c in existing
                    if c.commit_sequence >= commit.commit_sequence
                )
                for e in c
            ]
            if self._conflict_detector.conflicts_with(
                list(map(self._get_body, commit.events)),
                list(map(self._get_body, commits)),
            ):
                raise ConflictingCommitException("Conflicting commit")
            else:
                raise NonConflictingCommitException("Non-conflicting commit")
        existing.append(commit)

    @staticmethod
    def _get_body(em: EventMessage) -> BaseEvent:
        body: BaseEvent = em.body
        return body

    def _ensure_stream(self, tenant_id: str, stream_id: str) -> bool:
        if tenant_id not in self._buckets:
            self._buckets[tenant_id] = {stream_id: list()}
            return False
        if stream_id not in self._buckets[tenant_id]:
            self._buckets[tenant_id] = {stream_id: list()}
            return False
        return True



import datetime
import inspect
import logging
import uuid
from typing import Type, Dict

from pydantic_core import from_json

from aett.domain import Aggregate
from aett.domain.async_aggregate_repository import AsyncAggregateRepository
from aett.eventstore import (
    ICommitEventsAsync,
    IAccessSnapshotsAsync,
    MAX_INT,
    Commit,
    Snapshot,
)


class AsyncDefaultAggregateRepository(AsyncAggregateRepository):
    def __init__(
        self,
        tenant_id: str,
        store: ICommitEventsAsync,
        snapshot_store: IAccessSnapshotsAsync,
        logger: logging.Logger | None = None,
    ):
        """
        Initialize the default aggregate repository.

        :param tenant_id: The tenant id of the repository instance
        :param store: The event store to use
        :param snapshot_store: The snapshot store to use
        """
        self._tenant_id = tenant_id
        self._store = store
        self._snapshot_store = snapshot_store
        self._logger = (
            logger
            if logger is not None
            else logging.getLogger(AsyncDefaultAggregateRepository.__name__)
        )

    async def get(
        self,
        cls: Type[AsyncAggregateRepository.TAggregate],
        stream_id: str,
        max_version: int = MAX_INT,
    ) -> AsyncAggregateRepository.TAggregate:
        self._logger.debug(
            f"Getting aggregate {cls.__name__} with id {stream_id} at version {max_version}"
        )
        snapshot = await self._snapshot_store.get(
            tenant_id=self._tenant_id, stream_id=stream_id, max_revision=max_version
        )
        min_version = 0
        commit_sequence = 0
        if snapshot is not None:
            min_version = snapshot.stream_revision + 1
            commit_sequence = snapshot.commit_sequence
        commits = []
        ita = self._store.get(
            tenant_id=self._tenant_id,
            stream_id=stream_id,
            min_revision=min_version,
            max_revision=max_version,
        )
        async for x in ita:
            commits.append(x)
        if len(commits) > 0:
            commit_sequence = commits[-1].commit_sequence
        memento_type = (
            inspect.signature(cls.apply_memento).parameters["memento"].annotation
        )
        aggregate: Aggregate = cls(
            stream_id,
            commit_sequence,
            memento_type(**from_json(snapshot.payload))
            if snapshot is not None
            else None,
        )
        for commit in commits:
            for event in commit.events:
                aggregate.raise_event(event.body)
        aggregate.uncommitted.clear()
        return aggregate

    async def get_to(
        self,
        cls: Type[AsyncAggregateRepository.TAggregate],
        stream_id: str,
        max_time: datetime.datetime = datetime.datetime.max,
    ) -> AsyncAggregateRepository.TAggregate:
        self._logger.debug(
            f"Getting aggregate {cls.__name__} with id {stream_id} at time point {max_time:%Y%m%d-%H%M%S%z}"
        )
        ait = self._store.get_to(
            tenant_id=self._tenant_id, stream_id=stream_id, max_time=max_time
        )
        commits = [x async for x in ait]
        commit_sequence = commits[-1].commit_sequence if len(commits) > 0 else 0
        aggregate = cls(stream_id, commit_sequence, None)
        for commit in commits:
            for event in commit.events:
                aggregate.raise_event(event.body)
        aggregate.uncommitted.clear()
        return aggregate

    async def save(
        self,
        aggregate: AsyncAggregateRepository.TAggregate,
        headers: Dict[str, str] | None = None,
    ) -> None:
        self._logger.debug(
            f"Saving aggregate {aggregate.id} at version {aggregate.version}"
        )
        if headers is None:
            headers = {}
        if len(aggregate.uncommitted) == 0:
            return
        commit = Commit(
            tenant_id=self._tenant_id,
            stream_id=aggregate.id,
            stream_revision=aggregate.version,
            commit_id=uuid.uuid4(),
            commit_sequence=aggregate.commit_sequence + 1,
            commit_stamp=datetime.datetime.now(datetime.timezone.utc),
            headers=dict(headers),
            events=list(aggregate.uncommitted),
            checkpoint_token=0,
        )
        await self._store.commit(commit)
        self._logger.debug(f"Saved aggregate {aggregate.id}")
        aggregate.uncommitted.clear()

    async def snapshot(
        self,
        cls: Type[AsyncAggregateRepository.TAggregate],
        stream_id: str,
        version: int = MAX_INT,
        headers: Dict[str, str] | None = None,
    ) -> None:
        self._logger.debug(
            f"Snapshotting aggregate {cls.__name__} with id {stream_id} at version {version}"
        )
        agg = await self.get(cls, stream_id, version)
        await self._snapshot_aggregate(agg, headers)

    async def snapshot_at(
        self,
        cls: Type[AsyncAggregateRepository.TAggregate],
        stream_id: str,
        cut_off: datetime.datetime,
        headers: Dict[str, str] | None = None,
    ) -> None:
        self._logger.debug(
            f"Snapshotting aggregate {cls.__name__} with id {stream_id} at time point {cut_off:%Y%m%d-%H%M%S%z}"
        )
        agg = await self.get_to(cls, stream_id, cut_off)
        await self._snapshot_aggregate(agg, headers)

    async def _snapshot_aggregate(
        self, aggregate: Aggregate, headers: Dict[str, str] | None = None
    ) -> None:
        memento = aggregate.get_memento()
        snapshot = Snapshot(
            tenant_id=self._tenant_id,
            stream_id=aggregate.id,
            commit_sequence=aggregate.commit_sequence,
            payload=memento.model_dump_json(serialize_as_any=True),
            stream_revision=memento.version,
            headers=headers or {},
        )
        await self._snapshot_store.add(snapshot=snapshot, headers=headers)



from typing import Dict

from pydantic_core import from_json, to_json

from aett.eventstore import IAccessSnapshots, SNAPSHOTS, Snapshot, MAX_INT
from aett.storage.synchronous.s3 import S3Config


class SnapshotStore(IAccessSnapshots):
    def __init__(self, s3_config: S3Config, folder_name: str = SNAPSHOTS):
        self._s3_bucket = s3_config.bucket
        self._folder_name = folder_name
        self._resource = s3_config.to_client()

    def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        files = self._resource.list_objects(
            Bucket=self._s3_bucket,
            Delimiter="/",
            Prefix=f"{self._folder_name}/{tenant_id}/{stream_id}/",
        )
        if "Contents" not in files:
            return None
        keys = list(
            int(key.split("/")[-1].replace(".json", ""))
            for key in map(lambda r: r.get("Key"), files.get("Contents"))
            if int(key.split("/")[-1].replace(".json", "")) <= max_revision
        )
        keys.sort(reverse=True)

        key = f"{self._folder_name}/{tenant_id}/{stream_id}/{keys[0]}.json"
        j = self._resource.get_object(Bucket=self._s3_bucket, Key=key)
        d = from_json(j["Body"].read())
        return Snapshot(
            tenant_id=d.get("tenant_id"),
            stream_id=d.get("stream_id"),
            stream_revision=int(d.get("stream_revision")),
            commit_sequence=int(d.get("commit_sequence")),
            payload=d.get("payload"),
            headers=d.get("headers"),
        )

    def add(self, snapshot: Snapshot, headers: Dict[str, str] | None = None):
        if headers is not None:
            snapshot.headers.update(headers)
        key = f"{self._folder_name}/{snapshot.tenant_id}/{snapshot.stream_id}/{snapshot.stream_revision}.json"
        self._resource.put_object(
            Bucket=self._s3_bucket, Key=key, Body=to_json(snapshot)
        )



import typing

from aioboto3 import Session
from boto3.dynamodb.conditions import Key
from pydantic_core import from_json, to_json

from aett.storage.asynchronous.dynamodb import _get_client
from aett.eventstore import SNAPSHOTS, MAX_INT, Snapshot, IAccessSnapshotsAsync


class SnapshotStore(IAccessSnapshotsAsync):
    def __init__(
        self,
        table_name: str = SNAPSHOTS,
        region: str = "eu-central-1",
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        port: int = 8000,
    ):
        self.__session = Session(
            profile_name=profile_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region,
        )
        self.__port = port
        self.table_name = table_name

    async def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        try:
            async with _get_client(self.__session, self.__port) as client:
                table = await client.Table(self.table_name)
                query_response = await table.query(
                    TableName=self.table_name,
                    ConsistentRead=True,
                    Limit=1,
                    KeyConditionExpression=(
                        Key("TenantAndStream").eq(f"{tenant_id}{stream_id}")
                        & Key("StreamRevision").lte(max_revision)
                    ),
                    ScanIndexForward=False,
                )
                if len(query_response["Items"]) == 0:
                    return None
                item = query_response["Items"][0]
                return Snapshot(
                    tenant_id=item["TenantId"],
                    stream_id=item["StreamId"],
                    stream_revision=int(item["StreamRevision"]),
                    payload=item["Payload"],
                    commit_sequence=item["CommitSequence"],
                    headers=dict(from_json(item["Headers"])),
                )

        except Exception as e:
            raise Exception(
                f"Failed to get snapshot for stream {stream_id} with status code {e}"
            )

    async def add(
        self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None
    ):
        if headers is None:
            headers = {}
        try:
            item = {
                "TenantAndStream": f"{snapshot.tenant_id}{snapshot.stream_id}",
                "TenantId": snapshot.tenant_id,
                "StreamId": snapshot.stream_id,
                "StreamRevision": snapshot.stream_revision,
                "Payload": snapshot.payload,
                "CommitSequence": snapshot.commit_sequence,
                "Headers": to_json(headers).decode("utf-8"),
            }
            async with _get_client(self.__session, self.__port) as client:
                table = await client.Table(self.table_name)
                _ = await table.put_item(
                    TableName=self.table_name,
                    Item=item,
                    ReturnValues="NONE",
                    ReturnValuesOnConditionCheckFailure="NONE",
                    ConditionExpression="attribute_not_exists(TenantAndStream) AND attribute_not_exists(StreamRevision)",
                )
        except Exception as e:
            raise Exception(
                f"Failed to add snapshot for stream {snapshot.stream_id} with status code {e}"
            )



import typing

from aiomysql import connect
from pydantic_core import from_json, to_json

from aett.eventstore import IAccessSnapshotsAsync, SNAPSHOTS, MAX_INT, Snapshot


class AsyncSnapshotStore(IAccessSnapshotsAsync):
    def __init__(
        self,
        host: str,
        user: str,
        password: str,
        database: str,
        port: int = 3306,
        table_name: str = SNAPSHOTS,
    ):
        self._host: str = host
        self._user: str = user
        self._password: str = password
        self._database: str = database
        self._port: int = port
        self._table_name = table_name

    async def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        try:
            async with connect(
                host=self._host,
                user=self._user,
                password=self._password,
                db=self._database,
                port=self._port,
                autocommit=True,
            ) as connection:
                async with connection.cursor() as cur:
                    await cur.execute(
                        f"""SELECT *
          FROM {self._table_name}
         WHERE TenantId = %s
           AND StreamId = %s
           AND StreamRevision <= %s
         ORDER BY StreamRevision DESC
         LIMIT 1;""",
                        (tenant_id, stream_id, max_revision),
                    )
                    item = await cur.fetchone()
                    if item is None:
                        return None

                    return Snapshot(
                        tenant_id=item[0],
                        stream_id=item[1],
                        stream_revision=int(item[2]),
                        commit_sequence=int(item[3]),
                        payload=from_json(item[4]),
                        headers=dict(from_json(item[5])),
                    )
        except Exception as e:
            raise Exception(
                f"Failed to get snapshot for stream {stream_id} with error {e}"
            )

    async def add(
        self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None
    ):
        if headers is None:
            headers = {}
        try:
            async with connect(
                host=self._host,
                user=self._user,
                password=self._password,
                db=self._database,
                port=self._port,
                autocommit=True,
            ) as connection:
                async with connection.cursor() as cur:
                    await cur.execute(
                        f"""INSERT INTO {self._table_name} ( TenantId, StreamId, StreamRevision, CommitSequence, Payload, Headers) VALUES (%s, %s, %s, %s, %s, %s);""",
                        (
                            snapshot.tenant_id,
                            snapshot.stream_id,
                            snapshot.stream_revision,
                            snapshot.commit_sequence,
                            to_json(snapshot.payload),
                            to_json(headers),
                        ),
                    )
                await connection.commit()
        except Exception as e:
            raise Exception(
                f"Failed to add snapshot for stream {snapshot.stream_id} with error {e}"
            )



from typing import Iterable

import pymysql
from aett.eventstore import IManagePersistence, TopicMap, COMMITS, SNAPSHOTS, Commit
from aett.storage.synchronous.mysql import _item_to_commit


class PersistenceManagement(IManagePersistence):
    def __init__(
        self,
        host: str,
        user: str,
        password: str,
        database: str,
        topic_map: TopicMap,
        port: int = 3306,
        commits_table_name: str = COMMITS,
        snapshots_table_name: str = SNAPSHOTS,
    ):
        self.host: str = host
        self._user: str = user
        self._password: str = password
        self._database: str = database
        self._port: int = port
        self._topic_map = topic_map
        self._commits_table_name = commits_table_name
        self._snapshots_table_name = snapshots_table_name

    def initialize(self):
        try:
            stmts = [
                stmt.strip()
                for stmt in f"""CREATE TABLE IF NOT EXISTS {self._commits_table_name}
                     (
                         TenantId varchar(36) charset utf8 NOT NULL,
                         StreamId varchar(36) charset utf8 NOT NULL,
                         StreamIdOriginal varchar(1000) charset utf8 NOT NULL,
                         StreamRevision int NOT NULL CHECK (StreamRevision > 0),
                         Items tinyint NOT NULL CHECK (Items > 0),
                         CommitId binary(16) NOT NULL,
                         CommitSequence int NOT NULL CHECK (CommitSequence > 0),
                         CommitStamp TIMESTAMP NOT NULL,
                         CheckpointNumber bigint AUTO_INCREMENT,
                         Headers blob NULL,
                         Payload mediumblob NOT NULL,
                         PRIMARY KEY (CheckpointNumber)
                     );
                     CREATE UNIQUE INDEX IX_Commits ON {self._commits_table_name} (TenantId, StreamId, CommitSequence);
                     CREATE UNIQUE INDEX IX_Commits_CommitId ON {self._commits_table_name} (TenantId, StreamId, CommitId);
                     CREATE UNIQUE INDEX IX_Commits_Revisions ON {self._commits_table_name} (TenantId, StreamId, StreamRevision, Items);
                     CREATE INDEX IX_Commits_Stamp ON {self._commits_table_name} (CommitStamp);

                     CREATE TABLE IF NOT EXISTS {self._snapshots_table_name}
                     (
                         TenantId varchar(64) charset utf8 NOT NULL,
                         StreamId varchar(64) charset utf8 NOT NULL,
                         StreamRevision int NOT NULL CHECK (StreamRevision > 0),
                         CommitSequence int NOT NULL CHECK (CommitSequence > 0),
                         Payload blob NOT NULL,
                         Headers blob NOT NULL,
                         CONSTRAINT PK_Snapshots PRIMARY KEY (TenantId, StreamId, StreamRevision)
                     );""".split(";")
                if stmt.strip()
            ]
            with pymysql.connect(
                host=self.host,
                user=self._user,
                password=self._password,
                database=self._database,
                port=self._port,
                autocommit=True,
            ) as connection:
                with connection.cursor() as c:
                    for stmt in stmts:
                        c.execute(query=stmt)
                connection.commit()
        except Exception as e:
            print(e)
            pass

    def drop(self):
        with pymysql.connect(
            host=self.host,
            user=self._user,
            password=self._password,
            database=self._database,
            port=self._port,
            autocommit=True,
        ) as connection:
            with connection.cursor() as c:
                c.execute(f"""DROP TABLE {self._snapshots_table_name};""")
                c.execute(f"""DROP TABLE {self._commits_table_name};""")
            connection.commit()

    def purge(self, tenant_id: str):
        with pymysql.connect(
            host=self.host,
            user=self._user,
            password=self._password,
            database=self._database,
            port=self._port,
            autocommit=True,
        ) as connection:
            with connection.cursor() as c:
                c.execute(
                    f"""DELETE FROM {self._commits_table_name} WHERE TenantId = %s;""",
                    tenant_id,
                )
                c.execute(
                    f"""DELETE FROM {self._snapshots_table_name} WHERE TenantId = %s;""",
                    tenant_id,
                )
            connection.commit()

    def get_from(self, checkpoint: int) -> Iterable[Commit]:
        with pymysql.connect(
            host=self.host,
            user=self._user,
            password=self._password,
            database=self._database,
            port=self._port,
            autocommit=True,
        ) as connection:
            with connection.cursor() as cur:
                cur.execute(
                    f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                                  FROM {self._commits_table_name}
                                 WHERE CommitStamp >= %s
                                 ORDER BY CheckpointNumber;""",
                    (checkpoint,),
                )
                fetchall = cur.fetchall()
                for doc in fetchall:
                    yield _item_to_commit(doc, self._topic_map)



import sqlite3
import typing

from pydantic_core import from_json, to_json

from aett.eventstore import IAccessSnapshots, SNAPSHOTS, MAX_INT, Snapshot


class SnapshotStore(IAccessSnapshots):
    def __init__(self, connection_string: str, table_name: str = SNAPSHOTS):
        self._connection_string: str = connection_string
        self._table_name = table_name

    def get(
        self, tenant_id: str, stream_id: str, max_revision: int = MAX_INT
    ) -> Snapshot | None:
        try:
            connection = sqlite3.connect(self._connection_string)
            cur = connection.cursor()
            cur.execute(
                f"""SELECT *
          FROM {self._table_name}
          WHERE TenantId = ?
           AND StreamId = ?
           AND StreamRevision <= ?
         ORDER BY StreamRevision DESC
         LIMIT 1;""",
                (tenant_id, stream_id, max_revision),
            )
            item = cur.fetchone()
            if item is None:
                return None

            return Snapshot(
                tenant_id=item[0],
                stream_id=item[1],
                stream_revision=int(item[2]),
                commit_sequence=int(item[3]),
                payload=from_json(item[4]),
                headers={},
            )

        except Exception as e:
            raise Exception(
                f"Failed to get snapshot for stream {stream_id} with error {e}"
            )

    def add(self, snapshot: Snapshot, headers: typing.Dict[str, str] | None = None):
        try:
            with sqlite3.connect(self._connection_string) as connection:
                cur = connection.cursor()
                cur.execute(
                    f"""INSERT INTO {self._table_name} ( TenantId, StreamId, StreamRevision, CommitSequence, Payload) VALUES (?, ?, ?, ?, ?);""",
                    (
                        snapshot.tenant_id,
                        snapshot.stream_id,
                        snapshot.stream_revision,
                        snapshot.commit_sequence,
                        to_json(snapshot.payload),
                    ),
                )
                cur.close()
                connection.commit()

        except Exception as e:
            raise Exception(
                f"Failed to add snapshot for stream {snapshot.stream_id} with error {e}"
            )



from aett.storage.asynchronous.mongodb.async_commit_store import AsyncCommitStore
from aett.storage.asynchronous.mongodb.async_snapshot_store import AsyncSnapshotStore
from aett.storage.asynchronous.mongodb.async_persistence_management import (
    AsyncPersistenceManagement,
)
